import sys

from logging import debug, error, info, warning

import wx


def ShowQuestionDialog(frame, message, title="Question?", yes_no=False):
    exc_info = sys.exc_info()
    debug(message, exc_info=exc_info[0] and exc_info)
    if yes_no:
        flags = wx.YES | wx.NO | wx.YES_DEFAULT | wx.ICON_QUESTION
    else:
        flags = wx.OK | wx.CANCEL | wx.OK_DEFAULT | wx.ICON_QUESTION
    dlg = wx.MessageDialog(frame, message, title, flags)
    result = dlg.ShowModal() in [wx.ID_OK, wx.ID_YES]
    dlg.Destroy()
    return result


def ShowNotificationDialog(frame, message, title="Information"):
    exc_info = sys.exc_info()
    info(message, exc_info=exc_info[0] and exc_info)
    dlg = wx.MessageDialog(frame, message, title, wx.OK | wx.ICON_INFORMATION)
    dlg.ShowModal()
    dlg.Destroy()


def ShowWarningDialog(frame, message, title="Warning"):
    exc_info = sys.exc_info()
    warning(message, exc_info=exc_info[0] and exc_info)
    dlg = wx.MessageDialog(frame, message, title, wx.OK | wx.ICON_ERROR)
    dlg.ShowModal()
    dlg.Destroy()


def ShowErrorDialog(frame, message, title="ERROR"):
    exc_info = sys.exc_info()
    error(message, exc_info=exc_info[0] and exc_info)
    dlg = wx.MessageDialog(frame, message, title, wx.OK | wx.ICON_ERROR)
    dlg.ShowModal()
    dlg.Destroy()



"""
======================================================
:mod:`rigaku`  Rigaku ASCII file from Smartlab (*.ras)
======================================================

Loads the data exported by Rigaku Smartlab.
"""

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

try:
    import wx

    from wx.lib.masked import NumCtrl
except ImportError:

    class void:
        pass

    wx = void()
    wx.Dialog = void


class MeasurementCondition:
    operator = "unknown"
    system = "Unknown Rigaku"
    RASHeader = {}
    Axes = {}
    axis = "TwoThetaOmega"
    mode = "unknown"
    comment = ""
    memo = ""
    sample = ""
    file_type = "RAS_RAW"
    generator = {}
    detector = {}
    optics = {}
    auto_atten = False
    scan_unit = "deg"
    start_time = None

    # used to parse and represent the MeasurementCondition0.xml file content
    def __init__(self, root):
        for element in root:
            if element.tag == "RASHeader":
                self.parse_ras(element)
            elif element.tag == "GeneralInformation":
                self.parse_GI(element)
            elif element.tag == "ScanInformation":
                self.parse_scan(element)
            elif element.tag == "Axes":
                self.parse_axes(element)
            elif element.tag == "HWConfigurations":
                self.parse_hw(element)

    def parse_ras(self, element):
        RASHeader = {}
        for child in element:
            try:
                key, value = self.parse_pair(child)
            except ValueError:
                continue
            else:
                RASHeader[key] = value
        self.RASHeader = RASHeader

    def parse_GI(self, element):
        for child in element:
            if child.tag == "Operator":
                self.operator = child.text
            elif child.tag == "SystemName":
                self.system = child.text
            elif child.tag == "Comment":
                self.comment = child.text or ""
            elif child.tag == "Memo":
                self.memo = child.text or ""
            elif child.tag == "SampleName":
                self.sample = child.text

    def parse_scan(self, element):
        for child in element:
            if child.tag == "AxisName":
                self.axis = child.text
            elif child.tag == "Mode":
                self.mode = child.text
            elif child.tag == "AttenuatorAutoMode":
                self.auto_atten = child.text.strip().lower() == "true"
            elif child.tag == "PositionUnit":
                self.scan_unit = child.text
            elif child.tag == "StartTime":
                self.start_time = child.text

    def parse_axes(self, element):
        Axes = {}
        for child in element:
            try:
                key, value = self.parse_axis(child)
            except ValueError:
                continue
            else:
                Axes[key] = value
        self.Axes = Axes

    def parse_hw(self, element):
        for child in element:
            if child.tag == "XrayGenerator":
                generator = {}
                for sc in child:
                    generator[sc.tag] = sc.text
                self.generator = generator
            elif child.tag == "Detector":
                detector = {}
                for sc in child:
                    detector[sc.tag] = sc.text
                self.detector = detector
            elif child.tag == "Optics":
                optics = {}
                for sc in child:
                    optics[sc.tag] = sc.text
                self.optics = optics

    def parse_pair(self, element):
        if element.tag != "Pair":
            raise ValueError
        return element[0].text.lstrip("*"), element[1].text

    def parse_axis(self, element):
        if element.tag != "Axis":
            raise ValueError
        return element.attrib["Name"], dict(element.attrib)

    def to_dict(self):
        output = {}
        output["Generator"] = self.generator
        output["Detector"] = self.detector
        output["Optics"] = self.optics
        output["Axes"] = self.Axes
        output["RASHeader"] = self.RASHeader
        return output


class Plugin(Template):
    wildcard = "*.ras;*.rasx"

    def __init__(self, parent):
        Template.__init__(self, parent)
        self.x_col = 0
        self.y_col = 1
        self.e_col = 1
        self.xe_col = -1
        self.comment = "*"
        self.skip_rows = 0
        self.delimiter = None

    def ReadHeader(self, fh):
        header = {}
        hl = fh.readline()
        while hl.startswith("*"):
            if hl.startswith("*RAS_INT_START"):
                break
            if '"' in hl:
                keys, value = hl.split('"', 1)
                value = value.split('"')[0]
            else:
                keys = hl
                value = None
            hl = fh.readline()

            if value and "|" in value:
                value = value.split("|", 1)[-1]
            try:
                value = float(value)
            except (ValueError, TypeError):
                pass
            keys = keys[1:].lower().strip().split("_")
            if keys[0] not in header:
                header[keys[0]] = {}
            subhdr = header[keys.pop(0)]
            while len(keys) > 1:
                if keys[0] not in subhdr:
                    subhdr[keys[0]] = {}
                elif not isinstance(subhdr[keys[0]], dict):
                    subhdr[keys[0]] = {"_value": subhdr[keys[0]]}
                subhdr = subhdr[keys.pop(0)]
            subhdr[keys.pop(0)] = value
        del header["ras"]
        return header

    def get_from_header(self, header, keys):
        value = header.get(keys[0], None)
        if value is None or len(keys) == 1:
            return value
        else:
            if isinstance(value, dict):
                return self.get_from_header(value, keys[1:])
            else:
                return None

    def LoadData(self, dataset, filename, data_id=0):
        """
        Loads the data from filename into the data_item_number.
        """
        if filename.endswith(".rasx"):
            return self.LoadRasX(dataset, filename, data_id)
        try:
            with open(filename, encoding="utf-8", errors="ignore") as fh:
                header = self.ReadHeader(fh)
                load_array = np.loadtxt(fh, delimiter=self.delimiter, comments=self.comment, skiprows=self.skip_rows)
        except Exception as e:
            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n numpy.loadtxt"
                + " gave the following error:\n"
                + str(e),
            )
        else:
            dataset.x_raw = load_array[:, self.x_col]
            dataset.y_raw = load_array[:, self.y_col]
            dataset.error_raw = np.sqrt(load_array[:, self.y_col])
            if self.get_from_header(header, ["meas", "scan", "axis", "x", "_value"]) == "Omega/2-Theta":
                dataset.x_command = "2*x"
            else:
                dataset.x_command = "x"
            if load_array.shape[1] > 2 and self.get_from_header(header, ["hw", "r", "attenuater", "automode"]) == 1:
                # measurement uses attenuators
                dataset.set_extra_data("atten", load_array[:, 2])
                dataset.y_command = "y*atten"
                dataset.error_command = "e*atten"
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta["data_source"]["experiment"]["instrument"] = "RAS"
            dataset.meta["data_source"]["experiment"]["probe"] = "x-ray"
            dataset.meta["data_source"]["measurement"]["scheme"] = "angle-dispersive"
            inst = dataset.meta["data_source"]["measurement"]["instrument_settings"]
            inst["incident_angle"] = {
                "min": float(dataset.x_raw.min()),
                "max": float(dataset.x_raw.max()),
                "unit": self.get_from_header(header, ["meas", "scan", "unit", "x"]),
            }
            inst["wavelength"] = {
                "magnitude": self.get_from_header(header, ["hw", "xg", "wave", "length", "alpha1"]),
                "unit": "angstrom",
            }
            dataset.meta["data_source"]["sample"] = {"name": self.get_from_header(header, ["file", "sample"])}
            # insert all metadata into user field
            dataset.meta["ras_header"] = header

    def LoadRasX(self, dataset, filename, data_id):
        import io
        import xml.etree.ElementTree as ET
        import zipfile

        try:
            with zipfile.ZipFile(filename, "r") as zip_ref:
                profile_txt = zip_ref.open("Data0/Profile0.txt", "r").read()
                header_xml = zip_ref.open("Data0/MesurementConditions0.xml", "r").read()
            load_array = np.loadtxt(io.StringIO(profile_txt.decode("utf-8-sig")))
            info = MeasurementCondition(ET.fromstring(header_xml))
        except Exception as e:
            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n numpy.loadtxt / header analysis "
                + " gave the following error:\n"
                + str(e),
            )
        else:
            dataset.x_raw = load_array[:, 0]
            dataset.y_raw = load_array[:, 1]
            dataset.error_raw = np.sqrt(load_array[:, 1])
            if info.axis == "OmegaTwoTheta":
                dataset.x_command = "2*x"
            else:
                dataset.x_command = "x"
            if load_array.shape[1] > 2 and info.auto_atten:
                # measurement uses attenuators
                dataset.set_extra_data("atten", load_array[:, 2])
                dataset.y_command = "y*atten"
                dataset.error_command = "e*atten"
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta["data_source"]["owner"]["name"] = info.operator
            dataset.meta["data_source"]["experiment"]["instrument"] = info.system
            dataset.meta["data_source"]["experiment"]["probe"] = "x-ray"
            dataset.meta["data_source"]["experiment"]["title"] = info.memo + " - " + info.comment
            dataset.meta["data_source"]["experiment"]["start_date"] = info.start_time
            dataset.meta["data_source"]["measurement"]["scheme"] = "angle-dispersive"
            inst = dataset.meta["data_source"]["measurement"]["instrument_settings"]
            inst["incident_angle"] = {
                "min": float(dataset.x_raw.min()),
                "max": float(dataset.x_raw.max()),
                "unit": info.scan_unit,
            }
            inst["wavelength"] = {
                "magnitude": float(info.generator.get("WavelengthKalpha1", 1.540593)),
                "unit": "angstrom",
            }
            dataset.meta["data_source"]["sample"] = {"name": info.sample}
            # insert all metadata into user field
            dataset.meta["rasx_header"] = info.to_dict()



from functools import reduce

import numpy as np

from . import int_lay_xmean
from . import math_utils as mu

_ctype = np.complex128


def ass_P(k, d, ctype=_ctype):
    """Assamble the propagation matrix for the layers"""
    P = np.zeros((2, 2) + k.shape, dtype=ctype)
    P[0, 0] = np.exp(-1.0j * k * d)
    P[1, 1] = 1.0 / P[0, 0]
    return P


def ass_X(k, ctype=_ctype):
    """Assemble the interface matrix for all interfaces"""
    k_j = k[..., 1:]
    k_jm1 = k[..., :-1]

    X = np.empty((2, 2) + k_j.shape, dtype=ctype)

    X[0, 0] = (k_j + k_jm1) / k_jm1 / 2
    X[0, 1] = (k_jm1 - k_j) / k_jm1 / 2
    X[1, 0] = X[0, 1]
    X[1, 1] = X[0, 0]
    return X


def ass_X_2int(k_j, k_jm1, ctype=_ctype):
    """Assemble the interface matrix for all interfaces"""
    # k_j = k[...,1:]
    # k_jm1 = k[...,:-1]

    X = np.empty((2, 2) + k_j.shape, dtype=ctype)

    X[0, 0] = (k_j + k_jm1) / k_jm1 / 2
    X[0, 1] = (k_jm1 - k_j) / k_jm1 / 2
    X[1, 0] = X[0, 1]
    X[1, 1] = X[0, 0]
    return X


def ReflQ(Q, lamda, n, d, sigma):
    """Calculate the reflectivity from a multilayer stack with
    Abeles matrix formalism."""
    # pdb.set_trace()
    k_vac = 2 * np.pi / lamda
    kz = Q[:, np.newaxis] / 2.0
    n_amb2 = n[-1] * n[-1]
    k_j = np.sqrt((n * n - n_amb2) * k_vac * k_vac + n_amb2 * kz * kz)

    # print d.shape, n.shape, k_j.dtype
    X = ass_X(k_j)
    # X = im.ass_X_interfacelayer4(k_j, k_j, k_j, k_j*0, k_j*0, sigma, sigma*0, sigma*0)
    P = ass_P(k_j, d)

    PX = mu.dot2_Adiag(P[..., 1:-1], X[..., :-1])
    # print P.dtype, X.dtype, PX.dtype
    M = mu.dot2(X[..., -1], reduce(mu.dot2, np.rollaxis(PX, 3)[::-1]))

    r = M[1, 0] / M[0, 0]

    return np.abs(r) ** 2


def ReflQ_mag(Q, lamda, n, d, sigma, n_u, dd_u, sigma_u, n_l, dd_l, sigma_l):
    """Calculate the reflectivity from a multilayer stack with
    Abeles matrix formalism.

    The last layer is the substrate and the first layer is the ambient
    """
    # pdb.set_trace()
    k_vac = 2 * np.pi / (lamda * np.ones(Q.shape))[:, np.newaxis]
    kz = Q[:, np.newaxis] / 2.0
    n_amb2 = (n[:, 0] * n[:, 0])[:, np.newaxis]
    # print n_amb2.shape
    # print n.shape, n_l.shape, n_u.shape, kz.shape
    k = np.sqrt((n * n - n_amb2) * k_vac * k_vac + n_amb2 * kz * kz)
    k_l = np.sqrt((n_l * n_l - n_amb2) * k_vac * k_vac + n_amb2 * kz * kz)
    k_u = np.sqrt((n_u * n_u - n_amb2) * k_vac * k_vac + n_amb2 * kz * kz)
    # print 'k: ', k.shape, k_l.shape, k_u.shape
    # print 'sigma_l', sigma_l.shape
    # print 'dd_u ', dd_u.shape
    # print np.all(n_u == n), np.all(n_l == n)
    # print np.all(k_ju == k_j), np.all(k_jl == k_j)

    # print d.shape, n.shape, k_j.dtype
    # X = ass_X(k)
    # X = im.ass_X_interfacelayer4(k, k_u, k_l, dd_u, dd_l, sigma, sigma_l, sigma_u)
    X_lu = ass_X_2int(k_u[:, 1:], k_l[:, :-1])
    X_l = ass_X_2int(k_l[:, :-1], k[:, :-1])
    X_u = ass_X_2int(k[:, 1:], k_u[:, 1:])
    # print 'X: ', X_lu.shape, X_l.shape, X_u.shape

    X = int_lay_xmean.calc_iso_Xmean(X_l, X_lu, X_u, k, k_l, k_u, dd_u, dd_l, sigma, sigma_l, sigma_u)
    P = ass_P(k, d)

    PX = mu.dot2_Adiag(P[..., 1:-1], X[..., 1:])
    # print P.dtype, X.dtype, PX.dtype
    M = mu.dot2(X[..., 0], reduce(mu.dot2, np.rollaxis(PX, 3)))

    r = M[1, 0] / M[0, 0]

    return np.abs(r) ** 2


def ReflQ_ref(Q, lamda, n, d, sigma):
    # Length of k-vector in vaccum
    d = d[1:-1]
    sigma = sigma[:-1]
    Q0 = 4 * np.pi / lamda
    # Calculates the wavevector in each layer
    Qj = np.sqrt((n[:, np.newaxis] ** 2 - n[-1] ** 2) * Q0**2 + n[-1] ** 2 * Q**2)
    # Fresnel reflectivity for the interfaces
    rp = (Qj[1:] - Qj[:-1]) / (Qj[1:] + Qj[:-1])  # *exp(-Qj[1:]*Qj[:-1]/2*sigma[:,newaxis]**2)
    # print rp.shape #For debugging
    # print d.shape
    # print Qj[1:-1].shape
    p = np.exp(1.0j * d[:, np.newaxis] * Qj[1:-1])  # Ignoring the top and bottom layer for the calc.
    # print p.shape #For debugging
    # Setting up a matrix for the reduce function. Reduce only takes one array
    # as argument
    rpp = np.array(list(map(lambda x, y: [x, y], rp[1:], p)))

    # print rpp.shape
    # Paratt's recursion formula
    def formula(rtot, rint):
        return (rint[0] + rtot * rint[1]) / (1 + rtot * rint[0] * rint[1])

    # Implement the recursion formula
    r = reduce(formula, rpp, rp[0])
    # print r.shape
    # return the reflectivity
    return np.abs(r) ** 2



""" A library for handling optical constants and scattering lengths
in an efficent way. Based on a base class database which is subclassed
for each case.
Programmer: Matts Bjorck
Last changed: 2009-03-19
"""

import string

import numpy as np

# Imports needed for the dispersive table implementation
from scipy import interpolate

from genx.core.custom_logging import iprint

from . import refl_base as refl


# ==============================================================================
class Func(object):
    """A function object which stores the real function so it can be
    dynamically replaced from its parents.
    """

    def __init__(self, func):
        self.func = func

    def __call__(self, *args):
        return self.func(*args)

    def replace_func(self, func):
        self.func = func


class Proxy(object):
    """Proxy class borrowed from ASPN Recipe 496741 at adress
    http://code.activestate.com/recipes/496741/
    """

    __slots__ = ["_obj", "__weakref__"]

    def __init__(self, obj):
        object.__setattr__(self, "_obj", obj)

    #
    # proxying (special cases)
    #
    def __getattribute__(self, name):
        return getattr(object.__getattribute__(self, "_obj"), name)

    def __delattr__(self, name):
        delattr(object.__getattribute__(self, "_obj"), name)

    def __setattr__(self, name, value):
        setattr(object.__getattribute__(self, "_obj"), name, value)

    def __bool__(self):
        return bool(object.__getattribute__(self, "_obj"))

    def __str__(self):
        return str(object.__getattribute__(self, "_obj"))

    def __repr__(self):
        return repr(object.__getattribute__(self, "_obj"))

    def __call__(self, *args, **kwargs):
        return object.__getattribute__(self, "_obj").__call__(*args, **kwargs)

    def _change_object(self, obj):
        object.__setattr__(self, "_obj", obj)

    #
    # factories
    #
    _special_names = [
        "__abs__",
        "__add__",
        "__and__",
        "__cmp__",
        "__coerce__",
        "__contains__",
        "__delitem__",
        "__delslice__",
        "__div__",
        "__divmod__",
        "__eq__",
        "__float__",
        "__floordiv__",
        "__ge__",
        "__getitem__",
        "__getslice__",
        "__gt__",
        "__hash__",
        "__hex__",
        "__iadd__",
        "__iand__",
        "__idiv__",
        "__idivmod__",
        "__ifloordiv__",
        "__ilshift__",
        "__imod__",
        "__imul__",
        "__int__",
        "__invert__",
        "__ior__",
        "__ipow__",
        "__irshift__",
        "__isub__",
        "__iter__",
        "__itruediv__",
        "__ixor__",
        "__le__",
        "__len__",
        "__long__",
        "__lshift__",
        "__lt__",
        "__mod__",
        "__mul__",
        "__ne__",
        "__neg__",
        "__oct__",
        "__or__",
        "__pos__",
        "__pow__",
        "__radd__",
        "__rand__",
        "__rdiv__",
        "__rdivmod__",
        "__reduce__",
        "__reduce_ex__",
        "__repr__",
        "__reversed__",
        "__rfloorfiv__",
        "__rlshift__",
        "__rmod__",
        "__rmul__",
        "__ror__",
        "__rpow__",
        "__rrshift__",
        "__rshift__",
        "__rsub__",
        "__rtruediv__",
        "__rxor__",
        "__setitem__",
        "__setslice__",
        "__sub__",
        "__truediv__",
        "__xor__",
        "next",
    ]

    # @classmethod #This is the new way of doing a classmethod, starting from python 2.4
    def _create_class_proxy(cls, theclass):
        """creates a proxy for the given class"""

        def make_method(name):
            def method(self, *args, **kw):
                return getattr(object.__getattribute__(self, "_obj"), name)(*args, **kw)

            return method

        namespace = {}
        for name in cls._special_names:
            if hasattr(theclass, name) and not hasattr(cls, name):
                namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, theclass.__name__), (cls,), namespace)

    # And, this is the old way. Which I stck with for compatiblity reasons
    _create_class_proxy = classmethod(_create_class_proxy)

    def __new__(cls, obj, *args, **kwargs):
        """
        creates an proxy instance referencing `obj`. (obj, *args, **kwargs) are
        passed to this class' __init__, so deriving classes can define an
        __init__ method of their own.
        note: _class_proxy_cache is unique per deriving class (each deriving
        class must hold its own cache)
        """
        try:
            cache = cls.__dict__["_class_proxy_cache"]
        except KeyError:
            cls._class_proxy_cache = cache = {}
        try:
            theclass = cache[obj.__class__]
        except KeyError:
            cache[obj.__class__] = theclass = cls._create_class_proxy(obj.__class__)
        ins = object.__new__(theclass)
        # theclass.__init__(ins, obj, *args, **kwargs)
        return ins


def change_proxyobject(proxy, obj):
    """
    changes the current object of proxy to the obj.
    Note that evil things can happen if not care is taken so the objects are
    of the same class
    """
    object.__setattr__(proxy, "_obj", obj)


# ==============================================================================
class Database(object):
    """A database class where object memebers are dynamically acessed
    and stored. I.e. the constants are only looked up when needed.
    """

    def __init__(self):
        """__init__(self) --> None"""
        object.__setattr__(self, "stored_values", {})

    def __getattribute__(self, name):
        """__getattr__(self, name) --> object

        Looks up and returns the attribure name
        """
        if name.startswith("__"):
            return object.__getattribute__(self, name)
        name = name.lower()
        stored_values = object.__getattribute__(self, "stored_values")
        if name in stored_values:
            return stored_values[name]
        else:
            try:
                res = object.__getattribute__(self, "lookup_value")(name)
                if type(res) is complex:
                    res = ElementComplex(res, name)
                stored_values[name] = res
            except (LookupError, IOError) as e:
                raise LookupError("The name %s does not exist in the" "database" % name)
            return stored_values[name]

    def __setattr__(self, name, value):
        """__setattr__(self, name) --> None

        Just overlaoding the setattribute so a object cant be set.
        """
        raise TypeError("This object does not support assignements")

    def lookup_value(self, name):
        """lookup_value(self, name) --> object

        Used to (externally) lookup a value in a database to be inserted in
        local one for this object.
        """
        iprint("Looking up value")
        return 1

    def reset_database(self):
        """reset_database(self) --> None

        Resets the internal database
        """
        stored_values = object.__getattribute__(self, "stored_values")
        stored_values = {}


# ==============================================================================
class ElementComplex(complex):
    def __new__(cls, value, element):
        instance = super().__new__(cls, value)
        instance.composition = ((1.0, element),)
        return instance

    def __mul__(self, other):
        output = super().__mul__(other)
        if type(other) in [float, int]:
            output = super().__new__(ElementComplex, output)
            output.composition = tuple((value*other, element) for value, element in self.composition)
        return output

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        output = super().__mul__(other)
        if type(other) in [float, int]:
            output = super().__new__(ElementComplex, output)
            output.composition = tuple((value/other, element) for value, element in self.composition)
        return output

    def __add__(self, other):
        output = super().__add__(other)
        if type(other) is ElementComplex:
            output = super().__new__(ElementComplex, output)
            output.composition = self.composition + other.composition
        return output


class FormFactor(Database):
    """A database for the x-ray formfactor which includes the
    anomulous part as well as the angle dependent part. The object will
    return a function of sin(theta)/lambda
    """

    def __init__(self, wavelength, f_calc, weights=None):
        """__init__(self, wavelength, f_calc) --> None

        wavelength [float] the wavelength of the radiation in AA, f_calc a
        function which calculates (lookup the scattering factor)
        """
        Database.__init__(self)
        object.__setattr__(self, "wavelength", wavelength)
        object.__setattr__(self, "f_calc", f_calc)
        object.__setattr__(self, "_weights", weights)

    def set_wavelength(self, wavelength):
        """set_wavelength(self, wavelength) --> None

        sets the wavelength of the database. This will reset the
        all the values. I.e when called each value will be reloaded
        """
        # First check so we actually change the wavelength ...
        if abs(wavelength - object.__getattribute__(self, "wavelength")) > 1e-10:
            object.__setattr__(self, "wavelength", wavelength)
            # object.__getattribute__(self, 'reset_database')()
            stored_vals = object.__getattribute__(self, "stored_values")
            # for key in stored_vals:
            #    f = object.__getattribute__(self, 'lookup_value')(key)
            #    change_proxyobject(stored_vals[key], f)
            # Removing a bug with proxy adding does not work properly
            for key in stored_vals:
                stored_vals[key] = object.__getattribute__(self, "lookup_value")(key)

    def __getattribute__(self, name):
        """__getattribute__(self, name) --> value

        Finds and returns a attribute
        """
        if name == "set_wavelength" or name.startswith("__"):
            return object.__getattribute__(self, "set_wavelength")
        else:
            # print 'getting value'
            return Database.__getattribute__(self, name)

    def lookup_value(self, name):
        """lookup_value(self, name) --> f [function]

        looks up a value in the external database
        """
        weights = object.__getattribute__(self, "_weights")
        wl = object.__getattribute__(self, "wavelength")
        # f = Proxy((object.__getattribute__(self, 'f_calc')(name, wl)))
        # Removing a bug with proxy adding does not work properly
        f = object.__getattribute__(self, "f_calc")(name, wl)
        if isinstance(f, complex):
            if weights is None:
                f = ElementComplex(f, name)
            elif name in weights:
                f = ElementComplex(f, name)*(0.6022141/weights[name])
        return f


# ==============================================================================
class ScatteringLength(Database):
    """A simpler version of the database where everything is loaded at once
    good for simple and small databases. I.e the scatteringlength for
    a neutron which is largely wavelength independent.
    """

    def __init__(self, values):
        """__init__(self, values) --> None

        values is a dictonary of key value pairs that are used in the database.
        """
        Database.__init__(self)
        ele_vals = {}
        for key, value in values.items():
            if type(value) in [float, complex]:
                ele_vals[key] = ElementComplex(value, key)
        object.__setattr__(self, "stored_values", ele_vals)

    @classmethod
    def with_weights(cls, values, weights):
        """
        Create a new database the uses weight normalized scattering factors for use with g/cm³ density.
        """
        self = object.__new__(cls)
        Database.__init__(self)

        ele_vals = {}
        for key, value in values.items():
            if type(value) in [float, complex] and key in weights:
                ele_vals[key] = ElementComplex(value, key) * (0.6022141 / weights[key])
        object.__setattr__(self, "stored_values", ele_vals)
        return self

    def lookup_value(self, name):
        raise LookupError("The element %s does not exist in the database" % name)


# ==============================================================================
# Function to load databases and or values


def load_f0dabax(filename, create_rho=False):
    """load_dabax(filename) --> f0 (dictonary)

    loads a dabax file with f0 values and return a dictonary with
    f0(sin(theta)/lambda) for the element given by the key name.
    """

    def create_f(a, b, c):
        def f(sin_over_lambda):
            """f(sin_over_lambda) --> float

            The atomic form factor for x-rays. The non-dispertive part.
            """
            return c + sum(a[:, np.newaxis] * np.exp(-b[:, np.newaxis] * sin_over_lambda**2), 0)

        return f

    def create_rho(a, b, c):
        def rho(r, fp, wb, occ):
            return (
                sum(
                    a[:, np.newaxis]
                    / (8.0 * (np.pi * (b[:, np.newaxis] / 16 / np.pi**2 + wb)) ** 1.5)
                    * np.exp(-(r**2.0) / 4.0 / (b[:, np.newaxis] / 16 / np.pi**2 + wb)),
                    0,
                )
                + (c + fp) / (8.0 * (np.pi * wb) ** 1.5) * np.exp(-(r**2.0) / 4.0 / wb)
            ) * occ

        return rho

    with open(filename, "r") as f:
        real_label = ""
        temp_dict = {}
        for line in f.readlines():
            # Get the label for each line
            if line[0] == "#":
                label = line[1]
                ret = line[1:-1]
            else:
                label = "D"
                ret = line[:-1]
            # Gets the real label, atom name
            if label == "S":
                real_label = ret.split()[-1]
            # The row contains data
            if label == "D":
                temp_dict[real_label.lower()] = [float(x) for x in ret.split()]

    f0 = {}
    rho0 = {}
    for key in list(temp_dict.keys()):
        temp = temp_dict[key]
        # change the name to not have operators...
        if key[-1] == "-":
            key = key[:-1] + "m"
        if key[-1] == "+":
            key = key[:-1] + "p"
        f0[key] = create_f(np.array(temp[:4]), np.array(temp[5:]), temp[4])
        rho0[key] = create_rho(np.array(temp[:4]), np.array(temp[5:]), temp[4])
    if create_rho:
        return f0, rho0
    else:
        return f0


def create_fp_lookup(path):
    """create_f_lookup(filename) --> lookup_func(name, wavelength)

    Creates a lookup function to lookup element names and returns a function
    that yields dispersive scattering factors at Q = 0. NOTE wavelengths in AA
    """

    def lookup_func(name, wavelength):
        """lookup_func(name, wavelength) --> fp = f1 - 1.0J*f2

        looksup the scattering factors for a given wavelength given in AA for
        a element (note that the databases does not support ions).
        The data given is the dispersive part + f0 (non dispersive) at Q = 0.
        """
        e, f1, f2 = np.loadtxt(path + "%s.nff" % name.lower(), skiprows=1, unpack=True)
        energy = 1239.842 / wavelength * 10
        if energy >= e[-2] or energy <= e[1]:
            raise ValueError(
                "The energy/wavelength is outside the databse"
                + "range, the energy should be inside [%f,%f] " % (e[1], e[-2])
            )
        pos1 = np.argmin(abs(e - energy))
        # Is the energy point to the right or left of the current point
        # If it is ontop it doesn't matter since the interpolation will be exact at
        # the endpoints
        if (e[pos1] - energy) > 0:
            pos2 = pos1 - 1
        else:
            pos2 = pos1 + 1
        # A quick linear interpolation:
        f1_e = (energy - e[pos1]) * (f1[pos2] - f1[pos1]) / (e[pos2] - e[pos1]) + f1[pos1]
        f2_e = (energy - e[pos1]) * (f2[pos2] - f2[pos1]) / (e[pos2] - e[pos1]) + f2[pos1]

        return complex(f1_e - 1.0j * f2_e)

    return lookup_func


def create_bl_lookup(path, b_dict):
    """ """

    def lookup_func(name, wavelength):
        """lookup_func(name, wavelength) --> b = b_s - 1.0J*b_a

        looksup the energy dependent absorption cross sections
        and calculated the complex scattering length from it.
        """
        energy = (0.28601435 / wavelength) ** 2  # convert wavelength in Å to energy in eV

        if len(name) > 1 and name[0].lower() == "i" and name[1].isdigit():
            if name[-2].isdigit():
                ele = name[-1] + name[1:-1]
            else:
                ele = name[-2:] + name[1:-2]
        else:
            ele = name.capitalize()

        e, xs_s, xs_a = np.loadtxt(path + "%s.txt" % ele, unpack=True)

        if energy >= e[-2] or energy <= e[1]:
            raise ValueError(
                "The energy/wavelength is outside the databse"
                + "range, the energy should be inside [%f,%f] " % (e[1], e[-2])
            )
        pos1 = np.argmin(abs(e - energy))
        # Is the energy point to the right or left of the current point
        # If it is ontop it doesn't matter since the interpolation will be exact at
        # the endpoints
        if (e[pos1] - energy) > 0:
            pos2 = pos1 - 1
        else:
            pos2 = pos1 + 1
        # A quick linear interpolation:
        xs_si = (energy - e[pos1]) * (xs_s[pos2] - xs_s[pos1]) / (e[pos2] - e[pos1]) + xs_s[pos1]
        xs_ai = (energy - e[pos1]) * (xs_a[pos2] - xs_a[pos1]) / (e[pos2] - e[pos1]) + xs_a[pos1]

        breal = b_dict[name.lower()].real
        bimag = xs_ai / wavelength / 2.0 * 1e-3
        return breal - 1j * bimag

    return lookup_func


def create_fpdisp_lookup(path):
    """create_f_lookup(filename) --> lookup_func(name)

    Creates a lookup function to lookup element names and returns a function
    that yields a function of dispersive scattering factors f(E) at Q = 0. NOTE energy is in eV
    """

    def create_dispersion_func(name):
        e, f1, f2 = np.loadtxt(path + "%s.nff" % name.lower(), skiprows=1, unpack=True)
        f1interp = interpolate.interp1d(e, f1, kind="linear")
        f2interp = interpolate.interp1d(e, f2, kind="linear")

        def f(energy):
            return f1interp(energy) - 1.0j * f2interp(energy)

        return refl.ReflFunction(f, (np.mean(e),), {}, id="f(E)")

    return create_dispersion_func


def create_f_lookup(lookup_fp, f0):
    """create_f_lookup(lookup_fp, f0) --> lookup_func(name, wavelength)

    combines a f0 dictonary [dict] witht the lookup function for the
     anoumoulous scattering factors. Note that it is assumed that lookup_fp
    is a function on the form lookup_fp(name, wavelength) --> f1 - 1.0Jf2
    where name is the name of the element!
    """

    def lookup_func(name, wavelength):
        """Looks up the total angle dependent form factor
        f = f0 + f1 + 1.0J*f2 of element name. This dispersive part
        is independent on the ionicity but f0 not.
        """
        # Check if name corrspond to an ion
        element = None
        if len(name) > 1:
            if name[-2].isdigit():
                element = name[:-2]
            else:
                element = name
        else:
            element = name
        # Remove the non dispersive part for the given element
        fp = lookup_fp(element, wavelength) - f0[element](0)
        f0_name = f0[name]

        def f(sin_over_lambda):
            return f0_name(sin_over_lambda) + fp

        return f

    return lookup_func


def create_rho_lookup(lookup_fp, rho0, f0):
    """create_rho_lookup(lookup_fp, rho0, f0) --> lookup_func(name, wavelength)

    combines a rho0 dictonary [dict] with the lookup function for the
     anoumoulous scattering factors. Note that it is assumed that lookup_fp
    is a function on the form lookup_fp(name, wavelength) --> f1 - 1.0Jf2
    where name is the name of the element!
    """

    def lookup_func(name, wavelength):
        """Looks up the "electron density" of an atom as described by
        its formfactor, f = f0 + f1 + 1.0J*f2, of element name.
        The dispersive part is independent on the ionicity but f0 not.
        """
        # Check if name corrspond to an ion
        element = None
        if len(name) > 1:
            if name[-2].isdigit():
                element = name[:-2]
            else:
                element = name
        else:
            element = name
        # Remove the non dispersive part for the given element
        fp = lookup_fp(element, wavelength) - f0[element](0)
        rho0_name = rho0[name]

        def rho(r, wb, occ):
            return rho0_name(r, fp, wb, occ)

        return rho

    return lookup_func


def create_fw_lookup(lookup_fp, weight):
    """create_fw_lookup(lookup_fp, weight) --> lookup_func(name, wavelength)

    combines a f0 dictonary [dict] witht the lookup function for the
     anoumoulous scattering factors. Note that it is assumed that lookup_fp
    is a function on the form lookup_fp(name, wavelength) --> f1 - 1.0Jf2
    where name is the name of the element!
    """

    def lookup_func(name, wavelength):
        """Looks up the total form factor at Q = 0
        f = f1 + 1.0J*f2 of element name.
        """
        fp = lookup_fp(name, wavelength)
        w = weight[name]

        return fp / (w / 0.6022141)

    return lookup_func


def load_bdabax(filename):
    """load_bdabax(filename) --> b (dictonary)

    loads a dabax file with b (sld for neutrons) values
    and return a dictonary with b for the elements and isotopes
    given by the key name.
    """
    temp_dict = read_dabax(filename)
    # We are intrested in the scatttering lengths
    b_c = {}
    for key in temp_dict:
        val = temp_dict[key][1] + temp_dict[key][2] * 1.0j
        if key[0].isdigit():
            b_c["i" + key] = val
        else:
            b_c[key] = val

    return b_c


def load_fdabax(filename):
    """loads a dabax file with the scattering length tables returns a lookup
    function so that the wavelength can be changed."""
    table = read_dabax(filename)

    def lookup_func(name, wavelength):
        """Looks up the total form factor at Q = 0
        f = f1 + 1.0J*f2 of element name.
        """
        el = np.array(table[name])
        el = el.reshape(len(el) / 7, 7)
        e = el[:, 0]
        f1 = el[:, 1]
        f2 = el[:, 2]
        energy = 1239.842 / wavelength * 10 / 1e3  # keV
        if energy >= e[-2] or energy <= e[1]:
            raise ValueError(
                "The energy/wavelength is outside the databse"
                + "range, the energy should be inside [%f,%f] " % (e[1], e[-2])
            )
        pos1 = np.argmin(abs(e - energy))
        # Is the energy point to the right or left of the current point
        # If it is ontop it doesn't matter since the interpolation will be exact at
        # the endpoints
        if (e[pos1] - energy) > 0:
            pos2 = pos1 - 1
        else:
            pos2 = pos1 + 1
        # A quick linear interpolation:
        f1_e = (energy - e[pos1]) * (f1[pos2] - f1[pos1]) / (e[pos2] - e[pos1]) + f1[pos1]
        f2_e = (energy - e[pos1]) * (f2[pos2] - f2[pos1]) / (e[pos2] - e[pos1]) + f2[pos1]

        return f1_e - 1.0j * f2_e

    return lookup_func


def read_dabax(filename):
    """read_dabax_dict(filename) --> temp_dict [dict]

    Loads an entire dabax file to a dictonary. Scan names are the
    keys.
    """

    def tofloat(x):
        try:
            f = float(x)
        except:
            f = np.nan
        return f

    real_label = ""
    temp_dict = {}
    with open(filename) as f:
        for line in f.readlines():
            # Get the label for each line
            if line[0] == "#":
                label = line[1]
                ret = line[1:-1]
            else:
                label = "Data"
                ret = line[:-1]
            # Gets the real label, atom name
            if label == "S":
                real_label = ret.split()[-1]
            # The row contains data
            if label == "Data":
                # To get all values in the table
                if real_label.lower() not in temp_dict:
                    temp_dict[real_label.lower()] = [tofloat(x.split("(")[0]) for x in ret.split()]
                else:
                    temp_dict[real_label.lower()] += [tofloat(x.split("(")[0]) for x in ret.split()]

    return temp_dict


def load_atomic_weights_dabax(filename):
    """load__atomic_weights_dabax(filename) --> w (dictonary)

    loads a dabax file with w (sld for neutrons) values
    and return one dictonary with the atomic weights for the elements
    by the key name. Note that all isotopes are includes with the an
    i in front of them.
    """

    temp_dict = read_dabax(filename)
    # We are intrested in the scatttering lengths
    w_mean = {}
    for key in temp_dict:
        if key[0].isdigit():
            # This is an isotope
            w_mean["i" + key] = temp_dict[key][0]
            if len(temp_dict[key]) == 3:
                # The mean atomic mass for the element
                w_mean[key.lstrip(string.digits)] = temp_dict[key][2]

    return w_mean


def create_scatt_weight(scatt_dict, w_dict):
    """create_bw(scatt_dict, w_dict) --> sw_dict

    Makes a scattering length database for using with densities (g/cm3)
    """
    sw_dict = {}
    for key in scatt_dict:
        # print key, ' ',scatt_dict[key], ' ',w_dict[key]
        if key in w_dict:
            sw_dict[key] = scatt_dict[key] / complex(w_dict[key] / 0.6022141)
    return sw_dict



import cmath
import math

import numba

from numba import cuda
from numpy import complex128, float64, pi

if numba.__version__.split(".") > ["0", "55", "0"]:
    JIT_OPTIONS = dict(cache=True)
else:
    JIT_OPTIONS = {}


@cuda.jit(
    numba.void(
        numba.float64[:], numba.float64, numba.complex128[:], numba.float64[:], numba.float64[:], numba.float64[:]
    ),
    **JIT_OPTIONS
)
def ReflNB(theta, lamda, n, d, sigma, Rout):
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    ai = tx + ty * bw
    if ai >= theta.shape[0]:
        return

    layers = d.shape[0]
    n0 = n[-1]

    k = 2.0 * pi / lamda

    pre1 = 2.0 * n0 * k
    pre2 = pi / 180.0

    Qi = pre1 * cmath.sqrt((n[0] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)

    Qj = pre1 * cmath.sqrt((n[1] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
    # Fresnel reflectivity for the interfaces
    rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
    Aj = rpj
    Qi = Qj

    for lj in range(2, layers):
        Qj = pre1 * cmath.sqrt((n[lj] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

        pj = cmath.exp(1.0j * d[lj - 1] * Qi)
        part = Aj * pj
        Aj = (rpj + part) / (1.0 + part * rpj)

        Qi = Qj
    Rout[ai] = abs(Aj) ** 2


@cuda.jit(
    numba.void(
        numba.float64[:], numba.float64, numba.complex128[:], numba.float64[:], numba.float64[:], numba.complex128[:]
    ),
    **JIT_OPTIONS
)
def AmpNB(theta, lamda, n, d, sigma, Aout):
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    ai = tx + ty * bw
    if ai >= theta.shape[0]:
        return

    layers = d.shape[0]
    n0 = n[-1]

    k = 2.0 * pi / lamda

    pre1 = 2.0 * n0 * k
    pre2 = pi / 180.0

    Qi = pre1 * cmath.sqrt((n[0] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)

    Qj = pre1 * cmath.sqrt((n[1] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
    # Fresnel reflectivity for the interfaces
    rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
    Aj = rpj
    Qi = Qj

    for lj in range(2, layers):
        Qj = pre1 * cmath.sqrt((n[lj] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

        pj = cmath.exp(1.0j * d[lj - 1] * Qi)
        part = Aj * pj
        Aj = (rpj + part) / (1.0 + part * rpj)

        Qi = Qj
    Aout[ai] = Aj


def Refl(theta, lamda, n, d, sigma, return_int=True):
    threadsperblock = 32
    blockspergrid = (theta.shape[0] + (threadsperblock - 1)) // threadsperblock

    Ctheta = cuda.to_device(theta)
    Cn = cuda.to_device(n.astype(complex128))
    Cd = cuda.to_device(d)
    Csigma = cuda.to_device(sigma)

    if return_int:
        CRout = cuda.device_array(shape=theta.shape, dtype=float64)
        ReflNB[blockspergrid, threadsperblock](Ctheta, lamda, Cn, Cd, Csigma, CRout)
        return CRout.copy_to_host()
    else:
        CAout = cuda.device_array(shape=theta.shape, dtype=complex128)
        AmpNB[blockspergrid, threadsperblock](Ctheta, lamda, Cn, Cd, Csigma, CAout)
        return CAout.copy_to_host()


@cuda.jit(
    numba.void(
        numba.float64[:], numba.float64, numba.complex128[:], numba.float64[:], numba.float64[:], numba.float64[:]
    ),
    **JIT_OPTIONS
)
def ReflQNB(Q, Q0, n, d, sigma, Rout):
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    qi = tx + ty * bw
    if qi >= Q.shape[0]:
        return

    layers = d.shape[0]
    n0 = n[-1]

    Qi = cmath.sqrt((n[0] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)

    Qj = cmath.sqrt((n[1] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)
    # Fresnel reflectivity for the interfaces
    rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
    Aj = rpj
    Qi = Qj

    for lj in range(2, layers):
        Qj = cmath.sqrt((n[lj] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

        pj = cmath.exp(1.0j * d[lj - 1] * Qi)
        part = Aj * pj
        Aj = (rpj + part) / (1.0 + part * rpj)

        Qi = Qj
    Rout[qi] = abs(Aj) ** 2


@cuda.jit(
    numba.void(
        numba.float64[:], numba.float64, numba.complex128[:], numba.float64[:], numba.float64[:], numba.complex128[:]
    ),
    **JIT_OPTIONS
)
def AmpQNB(Q, Q0, n, d, sigma, Aout):
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    qi = tx + ty * bw
    if qi >= Q.shape[0]:
        return

    layers = d.shape[0]
    points = Q.shape[0]
    n0 = n[-1]

    Qi = cmath.sqrt((n[0] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)

    Qj = cmath.sqrt((n[1] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)
    # Fresnel reflectivity for the interfaces
    rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
    Aj = rpj
    Qi = Qj

    for lj in range(2, layers):
        Qj = cmath.sqrt((n[lj] ** 2 - n0**2) * Q0**2 + n0**2 * Q[qi] ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

        pj = cmath.exp(1.0j * d[lj - 1] * Qi)
        part = Aj * pj
        Aj = (rpj + part) / (1.0 + part * rpj)

        Qi = Qj
    Aout[qi] = Aj


def ReflQ(Q, lamda, n, d, sigma, return_int=True):
    threadsperblock = 32
    blockspergrid = (Q.shape[0] + (threadsperblock - 1)) // threadsperblock

    Q0 = 4.0 * pi / lamda

    CQ = cuda.to_device(Q)
    Cn = cuda.to_device(n.astype(complex128))
    Cd = cuda.to_device(d)
    Csigma = cuda.to_device(sigma)

    if return_int:
        CRout = cuda.device_array(shape=Q.shape, dtype=float64)
        ReflQNB[blockspergrid, threadsperblock](CQ, Q0, Cn, Cd, Csigma, CRout)
        return CRout.copy_to_host()
    else:
        CAout = cuda.device_array(shape=Q.shape, dtype=complex128)
        AmpQNB[blockspergrid, threadsperblock](CQ, Q0, Cn, Cd, Csigma, CAout)
        return CAout.copy_to_host()


@cuda.jit(
    numba.void(
        numba.float64[:], numba.float64[:], numba.complex128[:, :], numba.float64[:], numba.float64[:], numba.float64[:]
    ),
    **JIT_OPTIONS
)
def Refl_nvary2NB(theta, lamda, n, d, sigma, Rout):
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    ai = tx + ty * bw
    if ai >= theta.shape[0]:
        return

    layers = d.shape[0]

    pre2 = pi / 180.0

    n0 = n[-1, ai]
    ki = 2.0 * pi / lamda[ai]
    pre1 = 2.0 * n0 * ki

    Qi = pre1 * cmath.sqrt((n[0, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)

    Qj = pre1 * cmath.sqrt((n[1, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
    # Fresnel reflectivity for the interfaces
    rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
    Aj = rpj
    Qi = Qj

    for lj in range(2, layers):
        Qj = pre1 * cmath.sqrt((n[lj, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

        pj = cmath.exp(1.0j * d[lj - 1] * Qi)
        part = Aj * pj
        Aj = (rpj + part) / (1.0 + part * rpj)

        Qi = Qj
    Rout[ai] = abs(Aj) ** 2


@cuda.jit(
    numba.void(
        numba.float64[:],
        numba.float64[:],
        numba.complex128[:, :],
        numba.float64[:],
        numba.float64[:],
        numba.complex128[:],
    ),
    **JIT_OPTIONS
)
def Amp_nvary2NB(theta, lamda, n, d, sigma, Aout):
    # Thread id in a 1D block
    tx = cuda.threadIdx.x
    # Block id in a 1D grid
    ty = cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw = cuda.blockDim.x
    # Compute flattened index inside the array
    ai = tx + ty * bw
    if ai >= theta.shape[0]:
        return

    layers = d.shape[0]

    pre2 = pi / 180.0

    n0 = n[-1, ai]
    ki = 2.0 * pi / lamda[ai]
    pre1 = 2.0 * n0 * ki

    Qi = pre1 * cmath.sqrt((n[0, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)

    Qj = pre1 * cmath.sqrt((n[1, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
    # Fresnel reflectivity for the interfaces
    rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[0] ** 2)
    Aj = rpj
    Qi = Qj

    for lj in range(2, layers):
        Qj = pre1 * cmath.sqrt((n[lj, ai] / n0) ** 2 - math.cos(theta[ai] * pre2) ** 2)
        # Fresnel reflectivity for the interfaces
        rpj = (Qj - Qi) / (Qj + Qi) * cmath.exp(-Qj * Qi / 2.0 * sigma[lj - 1] ** 2)

        pj = cmath.exp(1.0j * d[lj - 1] * Qi)
        part = Aj * pj
        Aj = (rpj + part) / (1.0 + part * rpj)

        Qi = Qj
    Aout[ai] = Aj


def Refl_nvary2(theta, lamda, n, d, sigma, return_int=True):
    threadsperblock = 32
    blockspergrid = (theta.shape[0] + (threadsperblock - 1)) // threadsperblock

    Ctheta = cuda.to_device(theta)
    Clamda = cuda.to_device(lamda)
    Cn = cuda.to_device(n.astype(complex128))
    Cd = cuda.to_device(d)
    Csigma = cuda.to_device(sigma)

    if return_int:
        CRout = cuda.device_array(shape=theta.shape, dtype=float64)
        Refl_nvary2NB[blockspergrid, threadsperblock](Ctheta, Clamda, Cn, Cd, Csigma, CRout)
        return CRout.copy_to_host()
    else:
        CAout = cuda.device_array(shape=theta.shape, dtype=complex128)
        Amp_nvary2NB[blockspergrid, threadsperblock](Ctheta, Clamda, Cn, Cd, Csigma, CAout)
        return CAout.copy_to_host()



"""
==============
ParameterVault
==============

A plugin that allows the user to quickly store parameter grid values and the corresponding
simulation to compare different fits agains each other. 

Written by Artur Glavic
Last Changes 04/22/15
"""

import os
import sys

from copy import deepcopy

import wx

from wx.lib.mixins.listctrl import ListCtrlAutoWidthMixin

import genx.gui.images as img

from genx.parameters import Parameters

from .. import add_on_framework as framework
from ..utils import ShowInfoDialog, ShowQuestionDialog, ShowWarningDialog


class Plugin(framework.Template):
    _refplugin = None

    @property
    def refplugin(self):
        # check if reflectivity plugin is None or destoryed, try to connect
        if not self._refplugin:
            self._init_refplugin()
        return self._refplugin

    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.parent = parent

        parameter_panel = self.NewDataFolder("Vault")
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.parameter_panel = wx.Panel(parameter_panel)
        self.create_parameter_list()
        sizer.Add(self.parameter_panel, 1, wx.EXPAND | wx.GROW | wx.ALL)
        parameter_panel.SetSizer(sizer)
        parameter_panel.Layout()

        menu = self.NewMenu("P. Vault")
        self.mb_recalc_all = wx.MenuItem(
            menu,
            wx.NewId(),
            "Recalculate items",
            "Run simulations for each parameter set again and calculate FOM",
            wx.ITEM_NORMAL,
        )
        menu.Append(self.mb_recalc_all)
        self.mb_auto_store = wx.MenuItem(
            menu,
            wx.NewId(),
            "Save tables",
            "Save a parameters table file each time you add something to the vault",
            wx.ITEM_CHECK,
        )
        menu.Append(self.mb_auto_store)
        self.mb_load_all = wx.MenuItem(
            menu, wx.NewId(), "Recall saved tables", "Load all table parameters previously saved", wx.ITEM_NORMAL
        )
        menu.Append(self.mb_load_all)
        self.mb_auto_store.Check(False)

        self.parent.Bind(wx.EVT_MENU, self.OnRecalcAll, self.mb_recalc_all)
        self.parent.Bind(wx.EVT_MENU, self.OnAutoStore, self.mb_auto_store)
        self.parent.Bind(wx.EVT_MENU, self.OnLoadAll, self.mb_load_all)

    def _init_refplugin(self):
        ph = self.parent.plugin_control.plugin_handler
        if "Reflectivity" in ph.loaded_plugins:
            # connect to the reflectivity plugin for layer creation
            self._refplugin = ph.loaded_plugins["Reflectivity"]
        else:
            ShowWarningDialog(self.materials_panel, "Reflectivity plugin must be loaded", "Information")
            self._refplugin = None

    def create_toolbar(self):
        dpi_scale_factor = wx.GetApp().dpi_scale_factor
        tb_bmp_size = int(dpi_scale_factor * 20)
        size = wx.Size(tb_bmp_size, tb_bmp_size)

        self.bitmap_button_add = wx.BitmapButton(
            self.tool_panel,
            -1,
            wx.Bitmap(img.getaddImage().Scale(tb_bmp_size, tb_bmp_size)),
            size=size,
            style=wx.NO_BORDER,
        )
        self.bitmap_button_add.SetToolTip("Add current parameters to Vault")
        self.bitmap_button_delete = wx.BitmapButton(
            self.tool_panel,
            -1,
            wx.Bitmap(img.getdeleteImage().Scale(tb_bmp_size, tb_bmp_size)),
            size=size,
            style=wx.NO_BORDER,
        )
        self.bitmap_button_delete.SetToolTip("Delete selected parameter set")
        self.bitmap_button_apply = wx.BitmapButton(
            self.tool_panel,
            -1,
            wx.Bitmap(img.getstart_fitImage().Scale(tb_bmp_size, tb_bmp_size)),
            size=size,
            style=wx.NO_BORDER,
        )
        self.bitmap_button_apply.SetToolTip("Apply selected parameter set to the model")
        self.bitmap_button_plot = wx.BitmapButton(
            self.tool_panel,
            -1,
            wx.Bitmap(img.getplottingImage().Scale(tb_bmp_size, tb_bmp_size)),
            size=size,
            style=wx.NO_BORDER,
        )
        self.bitmap_button_plot.SetToolTip("Toggle plotting of the selected parameter set")

        space = (2, -1)
        self.sizer_hor = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer_hor.Add(self.bitmap_button_add, proportion=0, border=2)
        self.sizer_hor.Add(space)
        self.sizer_hor.Add(self.bitmap_button_delete, proportion=0, border=2)
        self.sizer_hor.Add(space)
        self.sizer_hor.Add(self.bitmap_button_apply, proportion=0, border=2)
        self.sizer_hor.Add(space)
        self.sizer_hor.Add(self.bitmap_button_plot, proportion=0, border=2)

        self.parameter_panel.Bind(wx.EVT_BUTTON, self.parameter_list.AddItem, self.bitmap_button_add)
        self.parameter_panel.Bind(wx.EVT_BUTTON, self.parameter_list.DeleteItem, self.bitmap_button_delete)
        self.parameter_panel.Bind(wx.EVT_BUTTON, self.parameter_list.ApplyParameters, self.bitmap_button_apply)
        self.parameter_panel.Bind(wx.EVT_BUTTON, self.parameter_list.TogglePlot, self.bitmap_button_plot)

    def create_parameter_list(self):
        """
        Create a list of materials and it's graphical representation as
        well as a toolbar.
        """
        self.tool_panel = wx.Panel(self.parameter_panel)
        self.parameter_list = ParameterList(self.parameter_panel, self)
        self.sizer_vert = wx.BoxSizer(wx.VERTICAL)
        self.parameter_panel.SetSizer(self.sizer_vert)

        self.create_toolbar()

        self.sizer_vert.Add(self.tool_panel, proportion=0, flag=wx.EXPAND, border=5)
        self.sizer_vert.Add((-1, 2))
        self.sizer_vert.Add(self.parameter_list, proportion=1, flag=wx.EXPAND, border=5)
        self.tool_panel.SetSizer(self.sizer_hor)

        ppanel = wx.Panel(self.parameter_panel)
        psizer = wx.GridSizer(2, 2, 0)
        ppanel.SetSizer(psizer)

        psizer.Add(wx.StaticText(ppanel, label="P1 index:"), proportion=0, flag=wx.EXPAND, border=2)
        self.P1_index = wx.SpinCtrl(ppanel)
        self.P1_index.SetRange(0, 100)
        psizer.Add(self.P1_index, proportion=1, flag=wx.EXPAND, border=2)

        psizer.Add(wx.StaticText(ppanel, label="P2 index:"), proportion=0, flag=wx.EXPAND, border=2)
        self.P2_index = wx.SpinCtrl(ppanel)
        self.P2_index.SetRange(0, 100)
        psizer.Add(self.P2_index, proportion=1, flag=wx.EXPAND, border=2)

        self.sizer_vert.Add(ppanel, proportion=0, flag=wx.EXPAND, border=5)

        self.parameter_panel.Bind(wx.EVT_SPINCTRL, self.parameter_list.UpdateParams, self.P1_index)
        self.parameter_panel.Bind(wx.EVT_SPINCTRL, self.parameter_list.UpdateParams, self.P2_index)

    def OnSimulate(self, event):
        """OnSimulate(self, event) --> None

        Updates stuff after simulation
        """
        colors = ["b", "r", "g", "c", "m", "y", "k"]
        styles = ["--", "-.", ":"]
        plot = self.parent.plot_data
        sld_plot = self.refplugin.sld_plot
        model = self.GetModel()
        k = 0
        for ignore, do_plot, params, data, slds in self.parameter_list.parameter_list:
            if do_plot and params != model.parameters.data:
                j = 0
                for i, di in enumerate(data):
                    if model.data[i].show:
                        plot.ax.plot(
                            di.x,
                            di.y_sim,
                            c=di.sim_color,
                            lw=di.sim_linethickness,
                            ls=styles[k % len(styles)],
                            marker=di.sim_symbol,
                            ms=di.sim_symbolsize,
                        )
                        # if len(slds)>i:
                        #     for key in slds[i]:
                        #         is_imag=key[:2]=='Im' or key[:4]=='imag'
                        #         if (is_imag and self.refplugin.show_imag_sld) or not is_imag:
                        #             if key!='z' and key!='SLD unit':
                        #                 sld_plot.ax.plot(slds[i]['z'], slds[i][key],
                        #                                       colors[j%len(colors)], ls=styles[k%len(styles)],
                        #                                       label=None)
                        #                 j+=1
                k += 1
        plot.AutoScale()
        plot.flush_plot()
        # sld_plot.flush_plot()

    def OnRecalcAll(self, event):
        pl = self.parameter_list
        model = self.GetModel()
        start_params = model.parameters.data
        for i in range(len(pl.parameter_list)):
            pl.Select(i, False)
        for i in range(len(pl.parameter_list)):
            pl.Select(i, True)
            if i > 0:
                pl.Select(i - 1, False)
            pl.ApplyParameters(None)
            model.simulate()
            slds = deepcopy(model.script_module.SLD)
            params = deepcopy(model.parameters.data)
            data = [di.copy() for di in model.data]
            pl.parameter_list[i] = [model.fom or 0.0, True, params, data, slds]
        pl.SetItemCount(len(pl.parameter_list))
        model.parameters.data = start_params
        self.OnAutoStore(None)

    def OnAutoStore(self, event):
        p = Parameters()
        base_name = self.GetModel().filename.rsplit(".", 1)[0]
        if self.mb_auto_store.IsChecked():
            for i, (ignore, ignore, params, ignore, ignore) in enumerate(self.parameter_list.parameter_list):
                p.data = params
                txt = p.get_ascii_output()
                open(base_name + "_%i.tab" % i, "w").write(txt)

    def OnLoadAll(self, event):
        pl = self.parameter_list
        base_name = self.GetModel().filename.rsplit(".", 1)[0]
        if not os.path.exists(base_name + "_%i.tab" % 0):
            return
        for ignore in range(len(pl.parameter_list)):
            pl.Select(0, True)
            pl.DeleteItem(None)
        i = 0
        p = Parameters()
        model = self.GetModel()
        start_params = model.parameters.data
        while os.path.exists(base_name + "_%i.tab" % i):
            txt = open(base_name + "_%i.tab" % i, "r").read()
            p.set_ascii_input(txt)
            model.parameters.data = deepcopy(p.data)
            model.simulate()
            pl.AddItem(None, simulate=False)
            i += 1
        model.parameters.data = start_params
        self.parent.eh_tb_simulate(None)


class ParameterList(wx.ListCtrl, ListCtrlAutoWidthMixin):
    """
    The ListCtrl for the materials data.
    """

    def __init__(self, parent, plugin):
        wx.ListCtrl.__init__(self, parent, -1, style=wx.LC_REPORT | wx.LC_VIRTUAL | wx.LC_EDIT_LABELS)
        ListCtrlAutoWidthMixin.__init__(self)
        self.plugin = plugin
        dpi_scale_factor = wx.GetApp().dpi_scale_factor
        if sys.platform.startswith("win"):
            font = wx.Font(
                9, wx.FONTFAMILY_SWISS, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, faceName="Lucida Sans Unicode"
            )
        else:
            font = wx.Font(
                9, wx.FONTFAMILY_ROMAN, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, encoding=wx.FONTENCODING_UNICODE
            )
        self.SetFont(font)

        # Set list length
        self.SetItemCount(0)

        self.parameter_list = []
        # Set the column headers
        for col, (text, width) in enumerate(
            [
                ("#", 24),
                ("FOM", 60),
                ("Show", 60),
                ("P1", 60),
                ("P2", 60),
            ]
        ):
            self.InsertColumn(col, text, width=width)

    def OnSelectionChanged(self, evt):
        if not self.toggleshow:
            indices = self._GetSelectedItems()
            indices.sort()
            if not indices == self.show_indices:
                self.data_cont.show_data(indices)
                self._UpdateData("Show data set flag toggled", data_changed=True)
                # Forces update of list control
                self.SetItemCount(self.data_cont.get_count())
        evt.Skip()

    def OnGetItemText(self, item, col):
        if col == 0:
            return str(item + 1)
        data = self.parameter_list[item]
        if col == 1:
            return "%.4g" % data[0]
        if col == 2:
            return str(data[1])
        if col == 3:
            try:
                idx = self.plugin.P1_index.GetValue()
                model = self.plugin.GetModel()
                idx = [i for i, p in enumerate(model.parameters.data) if p[2]][idx]
                return "%g" % data[2][idx][1]
            except:
                return ""
        if col == 4:
            try:
                idx = self.plugin.P2_index.GetValue()
                model = self.plugin.GetModel()
                idx = [i for i, p in enumerate(model.parameters.data) if p[2]][idx]
                return "%g" % data[2][idx][1]
            except:
                return ""

    def _GetSelectedItems(self):
        """_GetSelectedItems(self) --> indices [list of integers]
        Function that yields a list of the currently selected items
        position in the list. In order of selction, i.e. no order.
        """
        indices = [self.GetFirstSelected()]
        while indices[-1] != -1:
            indices.append(self.GetNextSelected(indices[-1]))

        # Remove the last will be -1
        indices.pop(-1)
        return indices

    def _CheckSelected(self, indices):
        """_CheckSelected(self, indices) --> bool
        Checks so at least data sets are selcted, otherwise show a dialog box
        and return False
        """
        # Check so that one dataset is selected
        if len(indices) == 0:
            ShowInfoDialog(self, "At least one data set has to be selected")
            return False
        return True

    def DeleteItem(self, ignore):
        index = self.GetFirstSelected()
        self.parameter_list.pop(index)
        self.SetItemCount(len(self.parameter_list))
        self.plugin.OnAutoStore(None)

    def AddItem(self, ignore, simulate=True):
        model = self.plugin.GetModel()
        slds = deepcopy(model.script_module.SLD)
        params = deepcopy(model.parameters.data)
        data = [di.copy() for di in model.data]
        self.parameter_list.append([model.fom or 0.0, True, params, data, slds])
        self.SetItemCount(len(self.parameter_list))
        self.plugin.OnAutoStore(None)
        if simulate:
            self.plugin.parent.eh_tb_simulate(None)

    def ApplyParameters(self, ignore):
        index = self.GetFirstSelected()
        model = self.plugin.GetModel()
        was_stored = False
        for ignore, ignore, params, ignore, ignore in self.parameter_list:
            if model.parameters.data == params:
                was_stored = True
                break
        if not was_stored:
            message = "Do you want to store the current parameters in the Vault" + " before applying the selected ones?"
            result = ShowQuestionDialog(self, message, "Store current parameters?")
            if result:
                self.AddItem(None)
        model.parameters.data = deepcopy(self.parameter_list[index][2])
        self.plugin.parent.paramter_grid._grid_changed()
        self.plugin.parent.paramter_grid.grid.ForceRefresh()
        self.plugin.parent.eh_tb_simulate(None)

    def TogglePlot(self, ignore):
        index = self.GetFirstSelected()
        self.parameter_list[index][1] = not self.parameter_list[index][1]
        self.RefreshItem(index)
        self.plugin.parent.eh_tb_simulate(None)

    def UpdateParams(self, ignore):
        if len(self.parameter_list) == 0:
            return
        self.RefreshItems(0, len(self.parameter_list) - 1)



from bumps import __version__ as bumps_version
if tuple(bumps_version.split('.'))<('1','0','3'):
    from bumps.fitproblem import nllf_scale
else:
    # Adapt to changed signature of nllf_scale function
    from bumps.fitproblem import nllf_scale as _nllf_scale
    def nllf_scale(problem, norm=True):
        dof = problem.dof if norm else 1
        npars = max(len(problem.getp()), 1)
        return _nllf_scale(dof, npars, norm=norm)
from bumps.formatnum import format_uncertainty
from bumps.monitor import TimedUpdate


class NBMonitor(TimedUpdate):
    """
    Display fit progress on the console
    """

    def __init__(self, problem, progress=0.25, improvement=5.0):
        TimedUpdate.__init__(self, progress=progress, improvement=improvement)
        from IPython.display import display
        from ipywidgets import HTML, HBox, Output, VBox

        self.problem = problem
        self.status_text = HTML()
        self.result_text = HTML()
        self.plot_out = Output()

        vbox = VBox([self.status_text, self.result_text])
        hbox = HBox([vbox, self.plot_out])
        display(hbox)
        self.steps = []
        self.chis = []

    def show_progress(self, history):
        scale, err = nllf_scale(self.problem)
        chisq = format_uncertainty(scale * history.value[0], err)
        self.status_text.value = (
            '<table width="50%%"><tr><td>step</td><td>%s</td><td>cost</td><td>%s</td></tr></table>'
            % (history.step[0], chisq)
        )
        self.steps.append(history.step[0])
        self.chis.append(scale * history.value[0])

    def show_improvement(self, history):
        from IPython.display import clear_output, display
        from matplotlib.pyplot import close, figure, plot

        p = self.problem.getp()
        try:
            self.problem.setp(history.point[0])
            out = '<table width="50%"><tr>'
            out += "</tr>\n<tr>".join(
                [
                    "<td>%s</td><td>%s</td><td>%s</td>" % (pi.name, pi.value, pi.bounds)
                    for pi in self.problem._parameters
                ]
            )
            self.result_text.value = out + "</td></tr></table>"
        finally:
            self.problem.setp(p)
        with self.plot_out:
            fig = figure()
            plot(self.steps, self.chis)
            clear_output()
            display(fig)
            close()



"""
Library for surface x-ray diffraction simulations of superlattices
==================================================================

The model is based on Fullertons algorithm for superlattices as
described in Phys. Rev. B vol. 45 p. 9292 (1992). 
"""

# Programmed by Matts Bjorck 20091215

import time

import numpy as np

from . import sxrd
from .lib.physical_constants import r_e
from .sxrd import AtomGroup, Instrument, Slab, SymTrans, UnitCell
from .utils import f, rho

__pars__ = ["SLSample", "SLStandard", "UnitCell", "Slab", "AtomGroup", "Instrument"]

__xlabel__ = "q-scan [r.l.u.]"
__ylabel__ = "Instnsity [a.u.]"


class SLSample:
    """Class that models a multilayer sample on top of a
    substrate according to Fullertons model as given in
    PRB ....
    """

    def __init__(self, inst, bulk_slab, superlattice, unitcell, bulk_sym=None):
        if bulk_sym is None:
            bulk_sym = []
        self.set_bulk_slab(bulk_slab)
        self.set_bulk_sym(bulk_sym)
        self.superlattice = superlattice
        self.inst = inst
        self.set_unit_cell(unitcell)

    def set_bulk_slab(self, bulk_slab):
        """Set the bulk unit cell to bulk_slab"""
        if type(bulk_slab) != type(sxrd.Slab()):
            raise TypeError("The bulk slab has to be a member of" " class Slab")
        self.bulk = bulk_slab

    def set_unit_cell(self, unit_cell):
        """Sets the unitcell of the sample"""
        if type(unit_cell) != type(sxrd.UnitCell(1.0, 1.0, 1.0)):
            raise TypeError("The bulk slab has to be a member" " of class UnitCell")
        if unit_cell is None:
            unit_cell = sxrd.UnitCell(1.0, 1, 0.0, 1.0)
        self.unit_cell = unit_cell

    def set_bulk_sym(self, sym_list):
        """Sets the list of allowed symmetry operations for the bulk

        sym_list has to be a list ([]) of symmetry elements from the
        class SymTrans
        """
        # Type checking
        if type(sym_list) != type([]):
            raise TypeError("The surface symmetries has to contained" " in a list")

        if not sym_list:
            sym_list = [sxrd.SymTrans()]

        if min([type(sym) == type(sxrd.SymTrans()) for sym in sym_list]) == 0:
            raise TypeError("All members in the symmetry list has to" " be a memeber of class SymTrans")

        self.bulk_sym = sym_list

    def calc_i(self, h, k, l):
        """Calculate the diffracted intensity from a superlattice.
        The diffracted intensity from the superlattice and the substrate
        are added. I.e. it is assumed that the films is not coherent with
        the substrate.
        """
        bulk_i = np.abs(self.calc_fb(h, k, l)) ** 2
        sl_i = np.abs(self.superlattice.calc_i(h, k, l))
        return (bulk_i + sl_i) * self.inst.inten

    def calc_fb(self, h, k, l):
        """Calculate the structure factors from the bulk"""
        dinv = self.unit_cell.abs_hkl(h, k, l)
        x, y, z, el, u, oc, c = self.bulk._extract_values()
        oc = oc / float(len(self.bulk_sym))
        f = sxrd._get_f(self.inst, el, dinv)
        # Calculate the "shape factor" for the CTRs
        eff_thick = self.unit_cell.c / np.sin(self.inst.alpha * np.pi / 180.0)
        alpha = r_e * self.inst.wavel * eff_thick / self.unit_cell.vol() * np.sum(f.imag, 1)
        denom = np.exp(2.0 * np.pi * 1.0j * l) * np.exp(-alpha) - 1.0
        # Delta functions to remove finite size effect in hk plane
        delta_funcs = (abs(h - np.round(h)) < 1e-12) * (abs(k - np.round(k)) < 1e-12)
        # Sum up the uc struct factors
        f_u = np.sum(
            oc
            * f
            * np.exp(-2 * np.pi**2 * u * dinv[:, np.newaxis] ** 2)
            * np.sum(
                [
                    np.exp(
                        2.0
                        * np.pi
                        * 1.0j
                        * (
                            h[:, np.newaxis] * sym_op.trans_x(x, y)
                            + k[:, np.newaxis] * sym_op.trans_y(x, y)
                            + l[:, np.newaxis] * z[np.newaxis, :]
                        )
                    )
                    for sym_op in self.bulk_sym
                ],
                0,
            ),
            1,
        )
        # Putting it all togheter
        fb = f_u / denom * delta_funcs

        return fb


class Superlattice:
    """Class that describe a superlattice, can be subclassed
    to implement different strain profiles, interdiffusion etc..
    """

    def __init__(self, inst, unit_cell, a_slab, b_slab, a_sym=None, b_sym=None):
        if a_sym is None:
            a_sym = []
        if b_sym is None:
            b_sym = []
        self.a_slab = a_slab
        self.b_slab = b_slab
        if not a_sym:
            self.a_sym = [sxrd.SymTrans()]
        else:
            self.a_sym = a_sym
        if not b_sym:
            self.b_sym = [sxrd.SymTrans()]
        else:
            self.b_sym = b_sym
        self.unit_cell = unit_cell
        self.inst = inst

    def _extract_slab_values(self, slabs, sym):
        """Extracts the necessary parameters for simulating
        a list of stacked slabs
        """
        # Extract the parameters we need
        # the star in zip(*... transform the list elements to arguments
        xt, yt, zt, elt, ut, oct, ct = list(zip(*[slab._extract_values() for slab in slabs]))

        x = np.r_[xt]
        y = np.r_[yt]
        # scale and shift the slabs with respect to each other
        cn = np.cumsum(np.r_[0, ct])[:-1]
        z = np.concatenate([zs * c_s + c_cum for zs, c_cum, c_s in zip(zt, cn, ct)])
        # el = reduce(lambda x,y:x+y, elt)
        el = np.r_[elt]
        u = np.r_[ut]
        oc = np.r_[oct]
        # print x,y,z, u
        t_lay = sum(ct)
        return x, y, z, u, oc, el, t_lay

    def calc_fslab(self, slablist, sym, h, k, l):
        """Calculate the structure factors from the bulk"""
        dinv = self.unit_cell.abs_hkl(h, k, l)
        x, y, z, u, oc, el, t_lay = self._extract_slab_values(slablist, sym)
        oc = oc / float(len(sym))
        f = sxrd._get_f(self.inst, el, dinv)
        # Sum up the uc struct factors
        f_u = np.sum(
            oc
            * f
            * np.exp(-2 * np.pi**2 * u * dinv[:, np.newaxis] ** 2)
            * np.sum(
                [
                    np.exp(
                        2.0
                        * np.pi
                        * 1.0j
                        * (
                            h[:, np.newaxis] * sym_op.trans_x(x, y)
                            + k[:, np.newaxis] * sym_op.trans_y(x, y)
                            + l[:, np.newaxis] * z[np.newaxis, :]
                        )
                    )
                    for sym_op in sym
                ],
                0,
            ),
            1,
        )
        # return f_u, (z.max() - z.min())*np.ones(l.shape)
        return f_u, t_lay * np.ones(l.shape)

    def calc_fa(self, n, h, k, l):
        """Calculate the strucutre factor for a a layer
        n is the thickness of the bilayer in units of slabs"""
        pass

    def calc_fb(self, n, h, k, l):
        """Calcualte the structure factor for a b layer
        n is the thickness of the bilayer in units of slabs"""
        pass

    def calc_fsl(self, unit_cell, h, k, l):
        """Calculate the strucutre factor for the entire
        superlattice.
        """
        raise NotImplementedError("calc_fsl has to be implemented in " "a Superlattices subclass")


class SLStandard(Superlattice):
    """Class that implements a "standard" superlattice, no strain
    included.
    """

    _pars = {"sigmaa": 1e-12, "sigmab": 1e-12, "repetitions": 2, "na": 2, "nb": 2, "a": 0.0, "c": 1e-12}

    def __init__(self, inst, unit_cell, a_slab, b_slab, a_sym=None, b_sym=None):
        Superlattice.__init__(self, inst, unit_cell, a_slab, b_slab, a_sym=a_sym, b_sym=b_sym)
        if a_sym is None:
            a_sym = []
        if b_sym is None:
            b_sym = []
        [self._make_set_func(name, self._pars[name]) for name in list(self._pars.keys())]
        [self._make_get_func(name) for name in list(self._pars.keys())]

    def calc_fa(self, n, h, k, l):
        f_slab, t_z = self.calc_fslab([self.a_slab] * n, self.a_sym, h, k, l)
        return f_slab, t_z

    def calc_fb(self, n, h, k, l):
        f_slab, t_z = self.calc_fslab([self.b_slab] * n, self.b_sym, h, k, l)
        return f_slab, t_z

    def thick_prob(self, n_mean, stand_dev):
        # According to fullerton it's enough to include three
        # standard deviations in the averaging
        lower = np.floor(n_mean - 3.0 * stand_dev)
        lower = np.int(min(lower, n_mean - 1))
        # We can't have a negative thickness, altough we remove
        # the gaussian distribution How does this affect the theoretical
        # assumptions?
        if lower < 1:
            lower = 1
        upper = np.ceil(n_mean + 3.0 * stand_dev)
        n = np.arange(lower, np.int(max(upper, n_mean + 1) + 1))
        # print 'n: ', n
        prob = np.exp(-((n - n_mean) ** 2) / 2.0 / stand_dev**2)
        prob = prob / sum(prob)
        return n, prob

    def calc_i(self, h, k, l):
        """Function to calculate the form factor from a superlattice"""
        # Create the different thicknesses to avarage over
        na, pa = self.thick_prob(self.na, self.sigmaa)
        nb, pb = self.thick_prob(self.nb, self.sigmab)
        tmp = np.array([self.calc_fa(n, h, k, l) for n in na])
        fa = tmp[:, 0, :]
        ta = tmp[:, 1, :]
        tmp = np.array([self.calc_fb(n, h, k, l) for n in nb])
        fb = tmp[:, 0, :]
        tb = tmp[:, 1, :]
        # print pa.shape, fa.shape
        pa = pa[:, np.newaxis]
        pb = pb[:, np.newaxis]
        # Do the different averagning
        fafa = (pa * fa * fa.conj()).sum(0)
        fbfb = (pb * fb * fb.conj()).sum(0)
        fam = (pa * fa).sum(0)
        fbm = (pb * fb).sum(0)
        phia = (pa * np.exp(2 * np.pi * 1.0j * ta * l) * fa.conj()).sum(0)
        phib = (pb * np.exp(2 * np.pi * 1.0j * tb * l) * fb.conj()).sum(0)
        ta = (pa * np.exp(2 * np.pi * 1.0j * ta * l)).sum(0)
        tb = (pb * np.exp(2 * np.pi * 1.0j * tb * l)).sum(0)
        m = self.repetitions
        ksi = 2 * np.pi * 1.0j * l * self.a - (2 * np.pi * l * self.c) ** 2 / 2.0
        # Calculate the intensity
        int = m * (fafa + 2.0 * np.real(np.exp(ksi) * phia * fbm) + fbfb) + 2.0 * np.real(
            (np.exp(-ksi) * phib * fam / ta / tb + phia * fam / ta + phib * fbm / tb + np.exp(ksi) * phia * fbm)
            * (
                (m - (m + 1) * np.exp(2.0 * ksi) * ta * tb + (np.exp(2.0 * ksi) * ta * tb) ** (m + 1))
                / (1 - np.exp(2.0 * ksi) * ta * tb) ** 2
                - m
            )
        )
        return int

    def _make_set_func(self, name, val):
        """Creates a function to set value for attribute with name"""

        def set_func(value):
            setattr(self, name, value)

        # Init the variable
        set_func(val)
        setattr(self, "set" + name, set_func)

    def _make_get_func(self, name):
        """Creates a get function"""

        def get_func():
            return getattr(self, name)

        setattr(self, "get" + name, get_func)


if __name__ == "__main__":
    from pylab import *

    inst = sxrd.Instrument(wavel=0.77, alpha=0.2)
    inst.set_inten(100.0)

    lay_a = sxrd.Slab()
    lay_a.add_atom("Sr", "sr", 0.0, 0.0, 0.0, 0.001, 1.0)
    lay_a.add_atom("Ti", "ti", 0.5, 0.5, 0.5, 0.001, 1.0)
    lay_a.add_atom("O1", "o", 0.5, 0.0, 0.5, 0.001, 1.0)
    lay_a.add_atom("O2", "o", 0.0, 0.5, 0.5, 0.001, 1.0)
    lay_a.add_atom("O3", "o", 0.5, 0.5, 0.0, 0.001, 1.0)

    lay_b = sxrd.Slab(c=1.0)
    lay_b.add_atom("La", "la", 0.0, 0.0, 0.0, 0.001, 1.0, 1.0)
    lay_b.add_atom("Al", "al", 0.5, 0.5, 0.5, 0.001, 1.0, 1.0)
    lay_b.add_atom("O1", "o", 0.5, 0.5, 0.0, 0.001, 1.0, 1.0)
    lay_b.add_atom("O2", "o", 0.0, 0.5, 0.5, 0.001, 1.0, 1.0)
    lay_b.add_atom("O3", "o", 0.5, 0.0, 0.5, 0.001, 1.0, 1.0)

    uc = sxrd.UnitCell(3.945, 3.945, 3.945, 90, 90, 90)

    sl = SLStandard(inst, uc, lay_b, lay_a)
    sl_sample = SLSample(inst, lay_a, sl, uc)
    sl.seta(0.0)
    sl.setc(0.00001)
    sl.setna(4.0)
    # Seems to have a lower limit of about 0.3 UC to work fine
    # with the calculation of thicknesses..
    sl.setsigmaa(0.3)
    sl.setnb(2.0)
    sl.setsigmab(0.3)
    sl.setrepetitions(10)
    l = np.arange(0.1, 3, 0.0011)
    h = 0.0 * np.ones(l.shape)
    k = 0.0 * np.ones(l.shape)
    int = sl_sample.calc_i(h, k, l)

    sample = sxrd.Sample(inst, lay_a, ([lay_b] * 4 + [lay_a] * 2) * 10, sxrd.UnitCell(3.945, 3.945, 3.945, 90, 90, 90))
    f_ref = sample.calc_f(h, k, l)
    int_ref = abs(f_ref) ** 2
    # Comparison between the normal sxrd model and the superlattice model.
    semilogy(l, int_ref)
    semilogy(l, int)
    legend(("sxrd model", "sxrd_mult closed form"))
    xlabel("l [r.l.u.]")
    ylabel("Intensity")
    show()



"""
Module to define all user exceptions raised by GenX.
"""


class GenxError(Exception):
    """Basic GenX error class to allow filtering for all specific GenX errors."""

    pass


class GenxIOError(IOError, GenxError):
    """Error class for input output, mostly concerning files"""

    def __init__(self, error_message, file=""):
        self.error_message = error_message
        self.file = file
        text = "Input/Output error for file:\n" + self.file + "\n\n Python error:\n " + self.error_message
        IOError.__init__(self, text)


class GenxOptionError(IndexError, GenxError):
    """Error class for not finding an option section pair in the configuration"""

    def __init__(self, section, option):
        self.section = section
        self.option = option
        text = (
            "Error in trying to loacate values in GenX configuration."
            + "\nCould not locate the section: "
            + self.section
            + " or option: "
            + self.option
            + "."
        )
        IndexError.__init__(self, text)


class ErrorBarError(RuntimeError, GenxError):
    def __init__(self, error_message=None):
        text = error_message
        if error_message is None:
            text = "Could not evaluate the error bars. A fit has to be run before they can be calculated"
        RuntimeError.__init__(self, text)


class ParameterError(ValueError, GenxError):
    """Class for yielding Parameter errors"""

    def __init__(self, parameter: str, parameter_number: int, py_error: Exception = None, what: int = -1):
        """
        parameter: the name of the parameter
        parameter_number: the position of the parameter in the list
        error_mesage: pythons error from the original exception
        int: where the error occured
            -1 : undefined
             0 : an not find the parameter
             1 : can not evaluate i.e. set the parameter
             2 : value are larger than max
             3 : value are smaller than min
             4 : No parameters to fit
        """
        self.parameter = parameter
        self.parameter_number = parameter_number
        self.what = what

        text = ""
        text += "Parameter number %i, %s, " % (self.parameter_number, self.parameter)

        # Take care of the different cases
        if self.what == 0:
            text += "could not be found. Check the spelling.\n"
        elif self.what == 1:
            text += "could not be evaluated. Check the code of the function.\n"
        elif self.what == 2:
            text += "is larger than the value in the max column.\n"
        elif self.what == 3:
            text += "is smaller than the value in the min column\n"
        elif self.what == 4:
            text = (
                "There are no parameter selcted to be fitted.\n"
                + "Select the parameters you want to fit by checking the "
                + "boxes in the fit column, folder grid"
            )
        else:
            text += "yielded an undefined error. Check the Python output\n"

        if py_error is not None:
            text += "\nPython error output:\n" + str(py_error)
        ValueError.__init__(self, text)


class ModelError(RuntimeError, GenxError):
    """Class for yielding compile or evaluation errors in the model text"""

    def __init__(self, error_message, where):
        """
        error_mesage: pythons error message from the original exception
        where: integer describing where the error was raised.
                -1: undef
                 0: compile error
                 1: evaulation error
        """
        self.where = where
        text = ""
        if self.where == 0:
            text += "It was not possible to compile the model script.\n"
        elif self.where == 1:
            text += "It was not possible to evaluate the model script.\n" + "Check the Sim function.\n"
        elif self.where == -1:
            text += "Undefined error from the Model. See below.\n"
        text += "\n" + error_message
        RuntimeError.__init__(self, text)


class FomError(RuntimeError, GenxError):
    """Error class for the fom evaluation"""

    def __init__(self, error_message):
        text = "Could not evaluate the FOM function. See python output.\n" + "\n" + error_message
        RuntimeError(self, text)


class OptimizerInterrupted(RuntimeError, GenxError):
    """Error raised to stop a running refinement function from within a thread"""



"""
=============
SpinAsymmetry
=============

A plugin to show an additional graph to plot the spin-asymmetry of polarized measurements.

Written by Artur Glavic
"""

import wx

from numpy import isnan

from genx.gui.plotpanel import BasePlotConfig, PlotPanel

from .. import add_on_framework as framework


class SAPlotConfig(BasePlotConfig):
    section = "spin asymmetry plot"


class SAPlotPanel(wx.Panel):
    """Widget for plotting the spin-asymmetry of datasets."""

    def __init__(self, parent, plugin, id=-1, color=None, dpi=None, style=wx.NO_FULL_REPAINT_ON_RESIZE, **kwargs):
        wx.Panel.__init__(self, parent)
        self.plot = PlotPanel(self, -1, color, dpi, SAPlotConfig, style, **kwargs)
        self.plugin = plugin

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.plot, 1, wx.EXPAND | wx.GROW | wx.ALL)

        self.plot.update(None)
        self.plot.ax = self.plot.figure.add_subplot(111)
        box = self.plot.ax.get_position()
        self.plot.ax.set_position([box.x0, box.y0, box.width * 0.95, box.height])
        self.plot.ax.set_autoscale_on(True)
        self.plot.update = self.Plot
        self.SetSizer(sizer)
        self.plot_dict = {}

    def SetZoom(self, active=False):
        return self.plot.SetZoom(active)

    def GetZoom(self):
        return self.plot.GetZoom()

    def Plot(self):
        """Plot(self) --> None

        Plotting the sample Sample.
        """
        colors = ["b", "r", "g", "c", "m", "y", "k"]
        # self.plot_dict = model.sample.SimSLD(None, model.inst)
        while len(self.plot.ax.lines) > 0:
            self.plot.ax.lines[0].remove()
        i = 0
        data = self.plugin.GetModel().get_data()
        sld_units = []

        # New style sim function with one sld for each simulation
        for i, di in enumerate(data):
            if i % 2 != 0:
                continue
            if di.show:
                try:
                    dj = data[i + 1]
                    li = len(di.x)
                    lj = len(dj.x)
                    l = min(li, lj)
                    SAdata = (di.y[:l] - dj.y[:l]) / (di.y[:l] + dj.y[:l])
                    if not isnan(SAdata).all():
                        self.plot.ax.plot(
                            di.x[:l],
                            SAdata,
                            color=di.data_color,
                            ls=di.data_linetype,
                            lw=di.data_linethickness,
                            marker=di.data_symbol,
                            ms=di.data_symbolsize,
                            label="data " + di.name + "&" + dj.name,
                        )
                    SAsim = (di.y_sim[:l] - dj.y_sim[:l]) / (di.y_sim[:l] + dj.y_sim[:l])
                    if not isnan(SAsim).all():
                        self.plot.ax.plot(
                            di.x[:l],
                            SAsim,
                            color=di.sim_color,
                            ls=di.sim_linetype,
                            lw=di.sim_linethickness,
                            marker=di.sim_symbol,
                            ms=di.sim_symbolsize,
                            label="sim " + di.name + "&" + dj.name,
                        )
                except:
                    pass

        self.plot.ax.legend(loc="upper right", framealpha=0.5, fontsize="small", ncol=1)  # bbox_to_anchor=(1, 0.5),

        self.plot.ax.yaxis.label.set_text("Spin Asymmetry")
        self.plot.ax.xaxis.label.set_text("x")
        wx.CallAfter(self.plot.flush_plot)
        self.plot.AutoScale()

    def ReadConfig(self):
        """ReadConfig(self) --> None

        Reads in the config file
        """
        return self.plot.ReadConfig()

    def GetYScale(self):
        """GetYScale(self) --> String

        Returns the current y-scale in use. Currently the string
        'log' or 'linear'. If the axes does not exist it returns None.
        """
        if self.plot.ax:
            return self.plot.ax.get_yscale()
        else:
            return None

    def GetXScale(self):
        """GetXScale(self) --> String

        Returns the current x-scale in use. Currently the string
        'log' or 'linear'. If the axes does not exist it returns None.
        """
        if self.plot.ax:
            return self.plot.ax.get_xscale()
        else:
            return None


class Plugin(framework.Template):
    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.parent = parent

        # Create the SLD plot
        SA_plot_panel = self.NewPlotFolder("Spin-Asymmetry")
        SA_sizer = wx.BoxSizer(wx.HORIZONTAL)
        SA_plot_panel.SetSizer(SA_sizer)
        self.SA_plot = SAPlotPanel(SA_plot_panel, self)
        SA_sizer.Add(self.SA_plot, 1, wx.EXPAND | wx.GROW | wx.ALL)
        SA_plot_panel.Layout()

    def OnSimulate(self, event):
        """OnSimulate(self, event) --> None

        Updates stuff after simulation
        """
        # Calculate and update the sld plot
        wx.CallAfter(self.SA_plot.Plot)

    def OnFittingUpdate(self, event):
        """OnSimulate(self, event) --> None

        Updates stuff during fitting
        """
        wx.CallAfter(self.SA_plot.Plot())



"""
Classes for advanced footprint correction to be used instead of the default square / gaussian beam profile.

All footprint classes follow the initial Footprint protocol to allow the model to evaluate them within the simulation.
"""

from dataclasses import dataclass

import numpy as np

from .base import ModelParamBase
from .instrument import GaussIntensity

__all__ = ["GaussianBeamOffset", "SquareBeamOffset", "TrapezoidBeam"]


class Footprint(ModelParamBase):
    """
    Base class for footprint corrections, needs to be sub-classed to provide additional functionality.

    Any subclass has to implement the __call__ method that applies its correction to an angle and sample length.
    """

    def __call__(self, theta, samplen):
        raise NotImplementedError("Subclass must implement __call__ method")


@dataclass
class GaussianBeamOffset(Footprint):
    """
    The footpring for a gaussian beam with sample off-centered.
    """

    sigma: float = 0.5
    offset: float = 0.0

    def __call__(self, theta, samplen):
        sinalpha = np.sin(theta * np.pi / 180.0)

        left = (samplen / 2.0) + self.offset / sinalpha
        right = (samplen / 2.0) - self.offset / sinalpha
        return GaussIntensity(theta, left, right, self.sigma)


@dataclass
class SquareBeamOffset(Footprint):
    """
    The footpring for a square beam with sample off-centered.
    Offset has to be smaller than width.
    """

    width: float = 0.1
    offset: float = 0.0

    def __call__(self, theta, samplen):
        sinalpha = np.sin(theta * np.pi / 180.0)
        cross_section = samplen * sinalpha
        c_max = cross_section / 2.0 + self.offset
        c_min = -cross_section / 2.0 + self.offset
        FP = np.maximum(0.0, (np.minimum(c_max, self.width / 2.0) - np.maximum(c_min, -self.width / 2.0))) / self.width
        return FP


@dataclass
class TrapezoidBeam(Footprint):
    """
    The footpring for a trapezoid beam. The intensity of the beam is constant
    on the inner_width and then reduces linearly until outer_width is reached.

    This footprint describes the typical neutron instrument situation, where the beam
    is defined by two slits, very well.
    """

    inner_width: float = 0.2
    outer_width: float = 0.3

    def __call__(self, theta, samplen):
        sinalpha = np.sin(theta * np.pi / 180.0)
        cross_section = samplen * sinalpha
        in_out_ratio = min(1.0, 2.0 * self.inner_width / (self.inner_width + self.outer_width))
        if self.inner_width > 0:
            FP_in = np.minimum(1.0, cross_section / self.inner_width)
        else:
            FP_in = 0.0
        if self.outer_width > self.inner_width:
            FP_out = np.minimum(
                1.0,
                np.maximum(
                    0.0,
                    1.0
                    - (
                        1.0
                        - np.minimum(1.0, (cross_section - self.inner_width) / (self.outer_width - self.inner_width))
                    )
                    ** 2,
                ),
            )
        else:
            FP_out = 0.0
        return FP_in * in_out_ratio + FP_out * (1.0 - in_out_ratio)



"""
Scripting interface for GenX for use in python scripting or Jupyter Notebooks.
"""

__all__ = []

import os
import sys

# workaround for issues with ctrl+c on windows
if sys.platform == "win32":
    os.environ["FOR_DISABLE_CONSOLE_CTRL_HANDLER"] = "1"

from .diffev import DiffEv
from .model import Model
from .model_control import ModelController
from .plugins.add_ons.help_modules.reflectivity_utils import SampleBuilder, SampleHandler, avail_models
from .plugins.utils import PluginHandler

_fit_output = []

controller = ModelController(DiffEv())


def text_output_api(text):
    _fit_output.append(text)


def load(fname, compile=True):
    controller.load_file(fname)
    if compile:
        controller.model.compile_script()
    controller.optimizer.model = controller.model
    return controller.model, controller.optimizer


def save(fname, model=None, optimizer=None):
    if model is not None:
        controller.model = model
    if optimizer is not None:
        controller.optimizer = optimizer
    controller.save_file(fname)


def fit_notebook(model, optimizer):
    """
    Function to fit a GenX model while giving feedback on a Jupyter notebook.
    """
    global _fit_output
    _fit_output = []
    optimizer.text_output = text_output_api
    optimizer.start_fit(model)
    from time import sleep

    import matplotlib.pyplot as plt

    from IPython.display import clear_output, display
    from numpy import array

    fig = plt.figure(figsize=(14, 5))
    plt.suptitle("To stop fit, interrupt the Kernel")

    plt.subplot(121)
    ax1 = plt.gca()
    line = plt.semilogy([0.0, 1.0], [0.1, 1.0])[0]
    plt.xlabel("Generation")
    plt.ylabel("FOM")
    t1 = plt.title("FOM:")

    plt.subplot(122)
    t2 = plt.title("Data display")
    ax2 = plt.gca()
    refls = []
    for i, ds in enumerate(model.data.items):
        refls.append(
            plt.semilogy(
                ds.x,
                ds.y,
                color=ds.data_color,
                lw=ds.data_linethickness,
                ls=ds.data_linetype,
                marker=ds.data_symbol,
                ms=ds.data_symbolsize,
                label="data-%i: %s" % (i, ds.name),
            )[0]
        )
        if ds.y_sim.shape == ds.y.shape:
            refls.append(
                plt.semilogy(
                    ds.x,
                    ds.y_sim,
                    color=ds.sim_color,
                    lw=ds.sim_linethickness,
                    ls=ds.sim_linetype,
                    marker=ds.sim_symbol,
                    ms=ds.sim_symbolsize,
                    label="model-%i: %s" % (i, ds.name),
                )[0]
            )

    plt.xlabel("x")
    plt.ylabel("I")

    plt.draw()

    display(fig)

    last = 2
    while optimizer.running:
        try:
            sleep(0.1)
            if len(optimizer.fom_log) <= last:
                continue
            x, y = array(optimizer.fom_log).T
            last = len(x)
            # t1.set_text('FOM: %.4e'%optimizer.best_fom)
            t1.set_text(_fit_output[-1])
            # vec=optimizer.best_vec
            # list(map(lambda func, value: func(value), model.get_fit_pars()[0], vec))
            # model.evaluate_sim_func()
            ax2.clear()
            t2 = plt.title("Data display")
            refls = []
            for i, ds in enumerate(model.data.items):
                refls.append(
                    plt.semilogy(
                        ds.x,
                        ds.y,
                        color=ds.data_color,
                        lw=ds.data_linethickness,
                        ls=ds.data_linetype,
                        marker=ds.data_symbol,
                        ms=ds.data_symbolsize,
                        label="data-%i: %s" % (i, ds.name),
                    )[0]
                )
                if ds.y_sim.shape == ds.y.shape:
                    refls.append(
                        plt.semilogy(
                            ds.x,
                            ds.y_sim,
                            color=ds.sim_color,
                            lw=ds.sim_linethickness,
                            ls=ds.sim_linetype,
                            marker=ds.sim_symbol,
                            ms=ds.sim_symbolsize,
                            label="model-%i: %s" % (i, ds.name),
                        )[0]
                    )

            plt.xlabel("x")
            plt.ylabel("I")

            line.set_xdata(x)
            line.set_ydata(y)

            ax1.set_xlim(0, x[-1])
            ax1.set_ylim(y.min() * 0.9, y.max() * 1.1)
            try:
                plt.draw()
            except Exception as e:
                print(e)
            else:
                clear_output(wait=True)
                display(fig)
        except KeyboardInterrupt:
            optimizer.stop_fit()
    plt.close()

    print(_fit_output[-1])
    print("If you want to update the model with the fit results, call api.fit_update(model, optimizer)")


def fit(model, optimizer):
    """
    Function to fit a GenX model while giving feedback with matplotlib graphs.
    """
    optimizer.text_output = print
    optimizer.start_fit(model)
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        plot_result = False
    else:
        plot_result = True
        from time import sleep

        from numpy import array

    if plot_result:
        fig = plt.figure(figsize=(14, 5))
        plt.ion()
        plt.suptitle("To stop fit, close the figure window")

        plt.subplot(121)
        ax1 = plt.gca()
        line = plt.semilogy([0.0, 1.0], [0.1, 1.0])[0]
        plt.xlabel("Generation")
        plt.ylabel("FOM")
        t1 = plt.title("FOM:")

        plt.subplot(122)
        t2 = plt.title("Data display")
        ax2 = plt.gca()
        refls = []
        for i, ds in enumerate(model.data.items):
            refls.append(
                plt.semilogy(
                    ds.x,
                    ds.y,
                    color=ds.data_color,
                    lw=ds.data_linethickness,
                    ls=ds.data_linetype,
                    marker=ds.data_symbol,
                    ms=ds.data_symbolsize,
                    label="data-%i: %s" % (i, ds.name),
                )[0]
            )
            if ds.y_sim.shape == ds.y.shape:
                refls.append(
                    plt.semilogy(
                        ds.x,
                        ds.y_sim,
                        color=ds.sim_color,
                        lw=ds.sim_linethickness,
                        ls=ds.sim_linetype,
                        marker=ds.sim_symbol,
                        ms=ds.sim_symbolsize,
                        label="model-%i: %s" % (i, ds.name),
                    )[0]
                )

        plt.xlabel("x")
        plt.ylabel("I")

        fig.canvas.mpl_connect("close_event", lambda event: optimizer.stop_fit())
        plt.draw()

    last = 2
    print("To stop fit, press ctrl+C")
    while optimizer.running:
        try:
            if plot_result:
                if len(optimizer.fom_log) <= last:
                    plt.pause(0.1)
                    continue
                x, y = array(optimizer.fom_log).T
                last = len(x)
                j = 0
                for i, ds in enumerate(model.data.items):
                    refls[j].set_ydata(ds.y)
                    if ds.y_sim.shape == ds.y.shape:
                        j = +1
                        refls[j].set_ydata(ds.y_sim)
                    j += 1

                line.set_xdata(x)
                line.set_ydata(y)

                ax1.set_xlim(0, x[-1])
                ax1.set_ylim(y.min() * 0.9, y.max() * 1.1)
                plt.draw()
            else:
                sleep(0.1)
        except KeyboardInterrupt:
            optimizer.stop_fit()
    if plot_result:
        plt.close()
    print("If you want to update the model with the fit results, call api.fit_update(model, optimizer)")


def fit_update(model, optimizer):
    pnames = model.parameters.get_fit_pars()[1]
    for di in model.parameters.data:
        if di[0] in pnames:
            di[1] = optimizer.best_vec[pnames.index(di[0])]
    model.simulate()


class Reflectivity(SampleBuilder):
    """
    Interface to build a model script for reflectivity simulations. Surves the same purpose
    as the Reflectivity plugin in the GUI.
    """

    _uvar_string = "cp.new_var("

    def __init__(self, model, optimizer, analyze_model=True):
        SampleBuilder.__init__(self, model)
        self.optimizer = optimizer
        if analyze_model:
            self.ReadModel()

    @classmethod
    def create_new(cls, modelname="spec_nx"):
        """
        Create a completely new GenX model for reflectivity.

        returns model, optimizer, Reflectivity
        """
        model = Model()
        optimizer = DiffEv()
        optimizer.model = model
        out = cls(model, optimizer, analyze_model=False)
        out.new_model(modelname)
        return model, optimizer, out

    def AppendSim(self, sim_func, inst, args):
        self.sim_funcs.append(sim_func)
        self.sim_insts.append(inst)
        self.sim_args.append(args)

    def new_model(self, modelname="spec_nx"):
        """Init the script in the model to yield the
        correct script for initilization
        """
        modelname = "models." + modelname
        model_data = self.GetModel().get_data()
        nb_data_sets = len(model_data)

        script = self.GetNewModelScript(modelname, nb_data_sets)
        self.BuildNewModel(script)

        self.instruments = {"inst": self.model.Instrument()}

        self.data_names = [data_set.name for data_set in model_data]
        self.uservars_lines = []
        self.sim_exp = [[] for item in self.data_names]
        self.sim_funcs = ["Specular"] * nb_data_sets
        self.sim_insts = ["inst"] * nb_data_sets
        self.sim_args = [["d.x"] for i in range(nb_data_sets)]
        self.WriteModel()

    def WriteModel(self):
        parameter_list = self.uservars_lines
        sim_funcs, sim_insts, sim_args = self.sim_funcs, self.sim_insts, self.sim_args
        expression_list = self.sim_exp
        instruments = self.instruments

        # make sure the number of datasets is reflected correctly
        model_data = self.GetModel().get_data()
        nb_data_sets = len(model_data)
        diff = nb_data_sets - len(sim_funcs)
        if diff > 0:
            expression_list += [[] for item in range(diff)]
            for i in range(diff):
                self.insert_new_data_segment(len(sim_funcs))
        elif diff < 0:
            for i in range(-diff):
                expression_list.pop(-1)
                self.remove_data_segment(nb_data_sets)

        self.write_model_script(sim_funcs, sim_insts, sim_args, expression_list, parameter_list, instruments)

    def update_script(self):
        self.WriteModel()

    def add_data(self, name=""):
        self.GetModel().data.add_new(name=name)
        self.update_script()

    def ReadModel(self):
        """ReadModel(self)  --> None

        Reads in the current model and locates layers and stacks
        and sample defined inside BEGIN Sample section.
        """
        self.CompileScript()

        instrument_names = self.find_instrument_names()

        if len(instrument_names) == 0:
            raise ValueError("Could not find any Instruments in the" " model script. Check the script.")

        if not "inst" in instrument_names:
            raise ValueError("Could not find the default Instrument, inst, in the" " model script. Check the script.")

        sample_text = self.find_sample_section()

        if sample_text is None:
            raise ValueError(
                "Could not find the sample section in the model script.\n " "Can not load the sample in the editor."
            )

        all_names, layers, stacks = self.find_layers_stacks(sample_text)

        if len(layers) == 0:
            raise ValueError("Could not find any Layers in the" " model script. Check the script.")

        # Now it's time to set all the parameters so that we have the strings
        # instead of the evaluated value - looks better
        from .models.lib.refl_base import ReflBase

        if isinstance(self.GetModel().script_module.sample, ReflBase):
            # TODO: revisit the whole method for new style model analysis
            pass
        else:
            for lay in layers:
                for par in lay[1].split(","):
                    vars = par.split("=")
                    exec(
                        '%s.%s = "%s"' % (lay[0], vars[0].strip(), vars[1].strip()),
                        self.GetModel().script_module.__dict__,
                    )

        data_names, insts, sim_args, sim_exp, sim_funcs = self.find_sim_function_parameters()

        uservars_lines = self.find_user_parameters()

        self.model = self.GetModel().script_module.model
        sample = self.GetModel().script_module.sample

        self.sampleh = SampleHandler(sample, all_names)
        self.sampleh.model = self.model
        instruments = {}
        for name in instrument_names:
            instruments[name] = getattr(self.GetModel().script_module, name)
        self.instruments = instruments

        self.data_names = data_names
        self.sim_exp = sim_exp
        self.uservars_lines = uservars_lines

        self.sim_funcs = sim_funcs
        self.sim_insts = insts
        self.sim_args = sim_args

        self.GetModel().compiled = False

    def add_stack(self, name, stack=None, **kwargs):
        if self.sampleh.checkName(name) or not name.isidentifier():
            raise KeyError("Name already exists or not proper identifier")
        self.sampleh.insertItem(len(self.sampleh) - 1, "Stack", name)
        stack = self.sampleh[name]
        if stack is not None:
            self[name] = stack
        else:
            for key, value in kwargs.items():
                if type(value) is str:
                    value = eval(value, self.GetModel().script_module.__dict__)
                setattr(stack, key, value)
            self.WriteModel()
        return self[name]

    def add_layer(self, name, layer=None, **kwargs):
        if self.sampleh.checkName(name) or name in self.instruments or not name.isidentifier():
            raise KeyError("Name already exists or not proper identifier")
        self.sampleh.insertItem(len(self.sampleh) - 1, "Layer", name)
        lay = self.sampleh[name]
        if layer is not None:
            self[name] = layer
        else:
            for key, value in kwargs.items():
                if type(value) is str:
                    value = eval(value, self.GetModel().script_module.__dict__)
                setattr(lay, key, value)
            self.WriteModel()
        return self[name]

    def add_instrument(self, name, instrument=None, **kwargs):
        if self.sampleh.checkName(name) or name in self.instruments or not name.isidentifier():
            raise KeyError("Name already exists or not proper identifier")
        if instrument is None:
            self.instruments[name] = self.Instrument(**kwargs)
            self.WriteModel()
        else:
            self.instruments[name] = instrument
        return self[name]

    def move_up(self, item):
        self.sampleh.moveUp(item)
        self.WriteModel()

    def move_down(self, item):
        self.sampleh.move_down(item)
        self.WriteModel()

    @property
    def Layer(self):
        return self.model.Layer

    @property
    def Stack(self):
        return self.model.Stack

    @property
    def Instrument(self):
        return self.model.Instrument

    def __getitem__(self, item):
        if item in self.sampleh.names:
            return self.sampleh[item]
        elif item in self.instruments:
            return self.instruments[item]
        else:
            raise KeyError("Has to be name of a Stack, Layer or Instrument")

    def __setitem__(self, key, value):
        if key in self.sampleh.names:
            prev = self.sampleh[key]
            if type(prev) != type(value):
                raise ValueError("Has to be of type %s" % type(prev))
            for k in prev._parameters.keys():
                setattr(prev, k, getattr(value, k))
        elif key in self.instruments:
            prev = self.instruments[key]
            if type(prev) != type(value):
                raise ValueError("Has to be of type %s" % type(prev))
            self.instruments[key] = value
        else:
            raise KeyError("Has to be name of a Stack, Layer or Instrument")
        self.WriteModel()

    def __repr__(self):
        output = "Reflectivity Model:\n"
        output += "    Instruments:"
        for key, value in self.instruments.items():
            output += "\n        %s = %s" % (key, value)

        output += "\n    Sample Structure:\n        "
        output += "\n        ".join(self.sampleh.getStringList(html_encoding=False))

        output += "\n    User Parameters:"
        for line in self.uservars_lines:
            var, val = line[len(self._uvar_string) :].split(")")[0].split(",")
            output += "\n        %s = %s" % (eval(var), val)

        output += "\n    Simulations:"
        for name, func, inst, exps in zip(self.data_names, self.sim_funcs, self.sim_insts, self.sim_exp):
            output += "\n        %s: %s(%s, %s)" % (name, func, "d.x", inst)
            for ei in exps:
                output += "\n            %s" % ei

        return output

    def _repr_html_(self):
        output = "<h3>Reflectivity Model</h3>"
        output += "<h4>Instruments:</h4>\n<div>"
        for key, value in self.instruments.items():
            output += self.sampleh.htmlize("%s = %s" % (key, value))
            output += "<br />"
        output += "</div>"

        output += "<h4>Sample Structure:</h4>\n<div>"
        output += "<br />\n".join(self.sampleh.getStringList(html_encoding=True))
        output += "</div>"

        output += "<h4>User Parameters:</h4>\n<div><ul>"
        for line in self.uservars_lines:
            var, val = line[len(self._uvar_string) :].split(")")[0].split(",")
            output += "<li><b>%s</b> = %s</li>" % (eval(var), val)
        output += "</ul></div>"

        output += "<h4>Simulations:</h4>\n<div>"
        for name, func, inst, exps in zip(self.data_names, self.sim_funcs, self.sim_insts, self.sim_exp):
            output += "<b>%s:</b> %s(%s, %s)" % (name, func, "d.x", inst)
            output += "<ul>"
            for ei in exps:
                output += "<li>%s</li>" % ei
            output += "</ul><br />"
        output += "</div>"

        return output

    @property
    def widget(self):
        # alternative widget to the model view with reflectivity controlls
        import ipywidgets as ipw

        model = self.GetModel()
        try:
            model.simulate()
        except Exception as error:
            print(error)

        graphw = ipw.Output()
        with graphw:
            from matplotlib import pyplot as plt

            fig = plt.figure(figsize=(10, 8))
            model.data.plot()
            plt.xlabel("q/tth")
            plt.ylabel("Intensity")
            from IPython.display import display

            display(fig)
            plt.close()

        dataw = model.data._repr_ipyw_()
        top = ipw.HBox([dataw, graphw])
        parameters = model.parameters._repr_ipyw_()

        replot = ipw.Button(description="simulate")
        replot.on_click(model._ipyw_replot)
        replot._plot_output = graphw

        tabs = ipw.Tab(children=[parameters, self.sample_widget, self.instrument_widget])
        tabs.set_title(0, "Parameters")
        tabs.set_title(1, "Sample")
        tabs.set_title(2, "Instrument")

        return ipw.VBox([replot, top, tabs])

    @property
    def instrument_widget(self):
        import ipywidgets as ipw

        entries = [ipw.HTML("<b>Instruments (use string to define formulas):</b>")]
        for key, value in self.instruments.items():
            ient = []
            if key == "inst":
                name = ipw.HTML(key, layout=ipw.Layout(width="14ex"))
            else:
                name = ipw.Text(value=key, layout=ipw.Layout(width="14ex"))
                name.observe(self._ipyw_change_iname, names="value")
            ient.append(name)
            ient.append(ipw.HTML("=Instrument(", layout=ipw.Layout(width="12ex")))

            items = str(value).split("(", 1)[1].rsplit(")", 1)[0].split(",")
            for item in items:
                try:
                    name, val = item.split("=", 1)
                except ValueError:
                    break
                name = name.strip()
                label_item = ipw.HTML("%s=" % name, layout=ipw.Layout(width="%iex" % (len(name) + 1)))
                ient.append(label_item)
                entry = ipw.Text(value=val, layout=ipw.Layout(width="18ex"))
                ient.append(entry)
                entry.change_item = name
                entry.change_name = key
                entry.label_item = label_item
                entry.observe(self._ipyw_change_eval, names="value")

            ient.append(ipw.HTML(")", layout=ipw.Layout(width="1ex")))
            entries.append(ipw.HBox(ient))
        btn = ipw.Button(description="Add Instrument")
        btn.on_click(self._ipyw_new_instrument)
        entries.append(btn)
        vbox = ipw.VBox(entries)
        btn.vbox = vbox
        return vbox

    @property
    def sample_widget(self):
        import ipywidgets as ipw

        entries = [ipw.HTML("<b>Sample (use string to define formulas):</b>")]

        maxlen = 200
        vbox = ipw.VBox([])

        for i, line in enumerate(self.sampleh.getStringList(html_encoding=False)):
            key, data = map(str.strip, line.split("=", 1))
            ient = []
            if key in ["Amb", "Sub"]:
                name_entr = ipw.HTML(key, layout=ipw.Layout(width="14ex"))
            else:
                name_entr = ipw.Text(value=key, layout=ipw.Layout(width="14ex"))
                name_entr.observe(self._ipyw_change_sname, names="value")
            ient.append(name_entr)

            if data.startswith("model.Stack"):
                ient.append(ipw.HTML("=Stack(", layout=ipw.Layout(width="12ex")))
                values = data[12:].strip()
            else:
                if key not in ["Amb", "Sub"]:
                    ient.insert(0, ipw.HTML("    ", layout=ipw.Layout(width="4ex")))
                cls, values = map(str.strip, data.split("(", 1))
                values = values.rsplit(")", 1)[0]
                ient.append(ipw.HTML("=%s(" % cls[6:], layout=ipw.Layout(width="12ex")))

            thislen = 14 + 12 + 4
            items = values.split(",")
            for item in items:
                try:
                    name, val = item.split("=", 1)
                except ValueError:
                    break
                name = name.strip()
                label_item = ipw.HTML("%s=" % name, layout=ipw.Layout(width="%iex" % (len(name) + 1)))
                ient.append(label_item)
                entry = ipw.Text(value=val, layout=ipw.Layout(width="18ex"))
                ient.append(entry)
                entry.change_item = name
                entry.change_name = key
                entry.label_item = label_item
                entry.observe(self._ipyw_change, names="value")
                thislen += 20 + (len(name) + 1)

            maxlen = max(maxlen, thislen + 1)
            ient.append(ipw.HTML(")", layout=ipw.Layout(width="1ex")))
            if data.startswith("model.Stack"):
                btn = ipw.Button(description="Add Layer")
                btn.on_click(self._ipyw_new_layer)
                btn.change_stack = key
                btn.vbox = vbox
                ient.append(btn)
            else:
                name_entr.input_widgets = tuple(wi for wi in ient[2:] if type(wi) is ipw.Text)
            entries.append(ipw.HBox(ient))

        btn = ipw.Button(description="Add Stack")
        btn.on_click(self._ipyw_new_stack)
        entries.append(btn)
        btn.vbox = vbox
        vbox.children = tuple(entries)
        vbox.layout.width = "%iex" % maxlen
        return vbox

    def _ipyw_change(self, change):
        item = self[change.owner.change_name]
        try:
            eval(change.new, self.GetModel().script_module.__dict__)
        except Exception as err:
            change.owner.label_item.value = '<div style="color: red;">%s=</div>' % change.owner.change_item
        else:
            change.owner.label_item.value = "%s=" % change.owner.change_item
            setattr(item, change.owner.change_item, change.new)
            self.WriteModel()

    def _ipyw_change_eval(self, change):
        item = self[change.owner.change_name]
        try:
            setattr(item, change.owner.change_item, eval(change.new, self.GetModel().script_module.__dict__))
        except Exception as err:
            change.owner.label_item.value = '<div style="color: red;">%s=</div>' % change.owner.change_item
        else:
            change.owner.label_item.value = "%s=" % change.owner.change_item
            self.WriteModel()

    def _ipyw_change_iname(self, change):
        self.instruments[change.new] = self.instruments[change.old]
        del self.instruments[change.old]
        for i, inst in enumerate(self.sim_insts):
            if inst == change.old:
                self.sim_insts[i] = str(change.new)
        self.WriteModel()

    def _ipyw_new_instrument(self, btn):
        key = "inst_%i" % len(self.instruments)
        self.add_instrument(key)
        value = self[key]

        import ipywidgets as ipw

        ient = []
        name = ipw.Text(value=key, layout=ipw.Layout(width="14ex"))
        name.observe(self._ipyw_change_iname, names="value")
        ient.append(name)
        ient.append(ipw.HTML("=Instrument(", layout=ipw.Layout(width="12ex")))

        items = str(value).split("(", 1)[1].rsplit(")", 1)[0].split(",")
        for item in items:
            try:
                name, val = item.split("=", 1)
            except ValueError:
                break
            name = name.strip()
            label_item = ipw.HTML("%s=" % name, layout=ipw.Layout(width="%iex" % (len(name) + 1)))
            ient.append(label_item)
            entry = ipw.Text(value=val, layout=ipw.Layout(width="18ex"))
            ient.append(entry)
            entry.change_item = name
            entry.change_name = key
            entry.label_item = label_item
            entry.observe(self._ipyw_change, names="value")

        ient.append(ipw.HTML(")", layout=ipw.Layout(width="1ex")))
        btn.vbox.children = btn.vbox.children[:-1] + (ipw.HBox(ient), btn.vbox.children[-1])
        self.WriteModel()

    def _ipyw_new_layer(self, btn):
        idx = [bi.children[-1] for bi in btn.vbox.children[1:-1]].index(btn) + 1
        key = "new_layer_%i" % len(self.sampleh.names)
        self.sampleh.insertItem(btn.change_stack, "Layer", key)
        lay = self[key]

        import ipywidgets as ipw

        ient = []
        name_entr = ipw.Text(value=key, layout=ipw.Layout(width="14ex"))
        name_entr.observe(self._ipyw_change_sname, names="value")
        ient.append(name_entr)
        data = repr(lay)

        ient.insert(0, ipw.HTML("    ", layout=ipw.Layout(width="4ex")))
        cls, values = map(str.strip, data.split("(", 1))
        values = values.rsplit(")", 1)[0]
        ient.append(ipw.HTML("=%s(" % cls, layout=ipw.Layout(width="12ex")))

        items = values.split(",")
        for item in items:
            try:
                name, val = item.split("=", 1)
            except ValueError:
                break
            name = name.strip()
            label_item = ipw.HTML("%s=" % name, layout=ipw.Layout(width="%iex" % (len(name) + 1)))
            ient.append(label_item)
            entry = ipw.Text(value=val, layout=ipw.Layout(width="18ex"))
            ient.append(entry)
            entry.change_item = name
            entry.change_name = key
            entry.label_item = label_item
            entry.observe(self._ipyw_change, names="value")
        ient.append(ipw.HTML(")", layout=ipw.Layout(width="1ex")))
        name_entr.input_widgets = tuple(wi for wi in ient[2:] if type(wi) is ipw.Text)

        btn.vbox.children = btn.vbox.children[: idx + 1] + (ipw.HBox(ient),) + btn.vbox.children[idx + 1 :]
        self.WriteModel()

    def _ipyw_change_sname(self, change):
        idx = self.sampleh.names.index(change.old)
        self.sampleh.names[idx] = change.new
        # update the name for the layer entries
        for entr in change.owner.input_widgets:
            entr.change_name = change.new
        self.WriteModel()

    def _ipyw_new_stack(self, btn):
        pass


class DataLoaderInterface:

    def __init__(self):
        head = os.path.dirname(os.path.abspath(__file__))
        self._handler = PluginHandler(None, os.path.join(head, "plugins", ""), "data_loaders")
        for dl in self._handler.get_possible_plugins():
            try:
                self._handler.load_plugin(dl)
                self._handler.loaded_plugins[dl].parent = self
                setattr(self, dl, self._handler.loaded_plugins[dl])
            except Exception as error:
                print(error)

    def SetStatusText(self, text):
        pass

    def __repr__(self):
        output = "Available data loaders:"
        for dl in sorted(self._handler.loaded_plugins.keys()):
            output += "\n  data_loader.%s.LoadData(dataset, filename)" % dl
        return output


data_loader = DataLoaderInterface()



"""
An interface to show output messages from python logging with convenient fielters.
"""

from typing import List

import wx

from ..model_actions import ActionBlock, ActionHistory, ModelAction
from .message_dialogs import ShowErrorDialog, ShowNotificationDialog


class ActionDisplayDialog(wx.Dialog):
    def __init__(self, parent, action: ModelAction):
        wx.Dialog.__init__(
            self, parent, style=wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP | wx.RESIZE_BORDER, name=f"GenX action details"
        )
        self.SetTitle(f"Action Details")
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        vbox.Add(wx.StaticText(self, label=f"Name: {action.action_name}"))
        vbox.Add(wx.StaticText(self, label=f"\nDescription:"))
        msg = wx.TextCtrl(self, value=str(action), style=wx.TE_READONLY | wx.TE_MULTILINE | wx.TE_DONTWRAP)
        font = msg.GetFont()
        font.SetFamily(wx.FONTFAMILY_TELETYPE)
        msg.SetFont(font)
        vbox.Add(msg, 1, wx.EXPAND)


class HistoryDialog(wx.Dialog):
    actions: List[ModelAction]
    current_index: int = 0
    changed_actions: ActionBlock

    def __init__(self, parent, history: ActionHistory):
        wx.Dialog.__init__(
            self, parent, style=wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP | wx.RESIZE_BORDER, name="ActionHistory"
        )
        self.SetTitle("GenX Action History")
        self.changed_actions = None
        self.history = history
        self.actions = history.undo_stack + list(reversed(history.redo_stack))
        self.current_index = len(history.undo_stack)
        self.build_layout()

        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.show_action_details, self.action_list)

        pos = parent.GetPosition()
        size = parent.GetSize()
        self.SetSize(400, int(size.height * 0.6))
        self.SetPosition(wx.Point(pos.x + size.width // 2, pos.y + int(size.height * 0.2)))

    def build_layout(self):
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)

        self.action_list = wx.ListCtrl(self, style=wx.LC_REPORT)
        self.action_list.AppendColumn("Action")
        self.append_actions()

        self.resize_columns()
        vbox.Add(self.action_list, 1, wx.EXPAND)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        vbox.Add(hbox)
        cbutton = wx.Button(self, -1, label="Close")
        hbox.Add(cbutton)
        self.Bind(wx.EVT_BUTTON, self.OnDestroy, cbutton)

        ubutton = wx.Button(self, -1, label="Revert selected actions (and remove from history)")
        self.Bind(wx.EVT_BUTTON, self.OnRevert, ubutton)
        hbox.Add(ubutton)

    def OnDestroy(self, evt):
        self.EndModal(wx.ID_CANCEL)

    def OnRevert(self, evt):
        start = self.action_list.GetFirstSelected()
        if 0 <= start < self.current_index:
            success = True
            items = [[start, 1]]
            nitem = self.action_list.GetNextSelected(start)
            while 0 <= nitem <= self.current_index:
                if nitem == sum(items[-1]):
                    items[-1][1] += 1
                else:
                    items.append([nitem, 1])
                nitem = self.action_list.GetNextSelected(nitem)
            changed: List[ModelAction] = []
            for start, length in items:
                try:
                    changed += self.history.remove_actions(self.current_index - start, length).actions
                except Exception as e:
                    ShowErrorDialog(
                        self,
                        f"The actions could not be re-applied:\n{e}\n\n"
                        f"The history was reset to the previous state. "
                        f"You might need to analyze your stepts in detail to revert these actions",
                    )
                    self.actions = self.history.undo_stack + list(reversed(self.history.redo_stack))
                    self.current_index = len(self.history.undo_stack)
                    changed += self.history.redo_stack[-length:]
                    self.action_list.DeleteAllItems()
                    self.append_actions()
                    success = False
                    break
            self.changed_actions = ActionBlock(self.actions[-1].model, changed)
            if success:
                self.EndModal(wx.ID_OK)
        else:
            ShowNotificationDialog(self, "You have to select at least one action to be removed from the undo stack")

    def append_actions(self):
        for i, action in enumerate(self.actions):
            self.action_list.Append((action.action_name,))
            if i >= self.current_index:
                self.action_list.SetItemTextColour(i, wx.Colour(100, 100, 100))

    def show_action_details(self, event):
        action = self.actions[event.GetIndex()]
        ActionDisplayDialog(self, action).ShowModal()

    def resize_columns(self):
        self.action_list.SetColumnWidth(0, wx.LIST_AUTOSIZE)
        self.action_list.EnsureVisible(self.action_list.GetItemCount() - 1)



"""
Classes for advanced resolution functions to be used instead of gaussian convolution.

All resolution classes follow the initial Resolution protocol to allow the model to evaluate them within the simulation.
"""

from dataclasses import dataclass
from typing import Union

import numpy as np

from .base import ModelParamBase
from .instrument import ResolutionVector

__all__ = ["GaussianResolution", "TrapezoidResolution", "XrayKalpha"]


class Resolution(ModelParamBase):
    """
    Base class for resolution calculations, needs to be sub-classed to provide additional functionality.

    Any subclass has to implement the __call__ method that applies its correction to an angle and sample length.
    The method returns a tuple of the 2 dimensional Q-positions for each calculation point and
    the weights for each. Later integration is done along the first axis. (See ResolutionVector for example.)
    """

    def __call__(self, TwoThetaQz, respoints=15, resintrange=2.0):
        raise NotImplementedError("Subclass must implement __call__ method")

    def get_weight_example(self):
        # Return a test sample of the resolution function for a q=0.1 and default parameters
        qres, weight = self(np.array([0.1]))
        weight = weight.flatten() / weight.max()
        return qres, weight.flatten()


@dataclass
class GaussianResolution(Resolution):
    sigma: Union[float, np.ndarray] = 0.005

    def __call__(self, TwoThetaQz, respoints=15, resintrange=2.0):
        # (TwoThetaQz, weight) =
        return ResolutionVector(TwoThetaQz[:], self.sigma, respoints, range=resintrange)


@dataclass
class TrapezoidResolution(Resolution):
    inner_width: Union[float, np.ndarray] = 0.02
    outer_width: Union[float, np.ndarray] = 0.0

    def __call__(self, TwoThetaQz, respoints=15, resintrange=2.0):
        full_width = np.maximum(self.inner_width, self.outer_width)
        scale = np.linspace(-0.5, 0.5, respoints)[:, np.newaxis]
        Qres = TwoThetaQz + full_width * scale
        # relative size of inner and outer widths, make sure the shape is like TwoThetaQz (widths can be float or array)
        inner_rel = self.inner_width / full_width * np.ones_like(TwoThetaQz)
        weight = np.where(2.0 * abs(scale) <= inner_rel, 1.0, 1.0 - (2.0 * abs(scale) - inner_rel) / (1.0 - inner_rel))
        return (Qres.flatten(), weight)

@dataclass
class XrayKalpha(Resolution):
    sigma_2theta: float
    energy_alpha1: float = 8047.78
    energy_alpha2: float = 8027.83
    ratio12: float = 2.0

    def __call__(self, TwoThetaQz, respoints=15, resintrange=2.0):
        """
        Create two gaussian resolution functions for k-alpha1 and k-alpha2 wavelengths.
        Assume that the wavelength defined in the model is for k-alpha1.
        """
        alpha2_qratio = self.energy_alpha2/self.energy_alpha1
        TTH1, weight1 = ResolutionVector(TwoThetaQz[:], self.sigma_2theta, respoints, range=resintrange)
        TTH2, weight2 = ResolutionVector(alpha2_qratio*TwoThetaQz[:], self.sigma_2theta, respoints, range=resintrange)
        I2 = 1.0/self.ratio12
        return np.vstack([TTH1.reshape(*weight1.shape), TTH2.reshape(*weight2.shape)]).flatten(), np.vstack([weight1, weight2*I2])
